from __future__ import annotations

import collections
import copy
import dataclasses
import traceback
from collections import defaultdict
from enum import Enum
from typing import Any, Collection, DefaultDict, Dict, List, Literal, Mapping, Optional, Tuple, Union

import pyarrow as pa

from chalk._gen.chalk.artifacts.v1 import export_pb2
from chalk._gen.chalk.artifacts.v1.export_pb2 import (
    MetadataSettings,
    ProjectSettings,
    ValidationLog,
    ValidationSettings,
)
from chalk._gen.chalk.graph.v1 import graph_pb2 as graph_pb
from chalk._gen.chalk.graph.v1.graph_pb2 import ResolverKind
from chalk._lsp.error_builder import FeatureClassErrorBuilder, LSPErrorBuilder, ResolverErrorBuilder
from chalk.features import FeatureConverter, FeatureSetBase
from chalk.features.pseudofeatures import PSEUDONAMESPACE
from chalk.features.resolver import RESOLVER_REGISTRY
from chalk.parsed._proto.utils import (
    RESOLVER_ENUM_TO_KIND,
    get_feature_reference_fqn,
    get_feature_type_attribute_name,
    get_feature_type_etl_offline_to_online,
    get_feature_type_fqn,
    get_feature_type_is_autogenerated,
    get_feature_type_max_staleness,
    get_feature_type_name,
    get_feature_type_namespace,
)
from chalk.utils.string import add_quotes, oxford_comma_list, to_camel_case


@dataclasses.dataclass
class ProtoField:
    name: str
    has_presence: bool


_METADATA_STRING_TO_FEATURE_ATTRIBUTE_MAP: Dict[str, ProtoField] = {
    "owner": ProtoField(name="owner", has_presence=True),
    "tags": ProtoField(name="tags", has_presence=False),
    "description": ProtoField(name="description", has_presence=True),
}

_METADATA_STRING_TO_RESOLVER_ATTRIBUTE_MAP: Dict[str, ProtoField] = {
    "owner": ProtoField(name="owner", has_presence=True),
    "tags": ProtoField(name="tags", has_presence=False),
    "description": ProtoField(name="doc", has_presence=True),
    "doc": ProtoField(name="doc", has_presence=True),
    "docstring": ProtoField(name="doc", has_presence=True),
}


class GraphLogSeverity(str, Enum):
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"


MissingMetadata = Dict[GraphLogSeverity, List[str]]


@dataclasses.dataclass
class MissingMetadataFeature:
    feature: graph_pb.FeatureType
    missing_metadata: MissingMetadata
    namespace_path: str


@dataclasses.dataclass
class MissingMetadataResolver:
    resolver: graph_pb.Resolver
    missing_metadata: MissingMetadata


class _LogBuilder:
    _severity_to_enum = {
        GraphLogSeverity.ERROR.value: export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        GraphLogSeverity.WARNING.value: export_pb2.VALIDATION_LOG_SEVERITY_WARNING,
        GraphLogSeverity.INFO.value: export_pb2.VALIDATION_LOG_SEVERITY_INFO,
    }

    def __init__(self):
        super().__init__()
        self._logs: List[ValidationLog] = []

    def add(self, log: ValidationLog):
        self._logs.append(log)

    def add_log(self, header: str, subheader: str, severity: GraphLogSeverity):
        if severity not in self._severity_to_enum:
            raise ValueError(f"Unknown severity: {severity}")
        self._logs.append(ValidationLog(header=header, subheader=subheader, severity=self._severity_to_enum[severity]))

    def add_error(self, header: str, subheader: str):
        self.add_log(header=header, subheader=subheader, severity=GraphLogSeverity.ERROR)

    def add_warning(self, header: str, subheader: str):
        self.add_log(header=header, subheader=subheader, severity=GraphLogSeverity.WARNING)

    def add_info(self, header: str, subheader: str):
        self.add_log(header=header, subheader=subheader, severity=GraphLogSeverity.INFO)

    def get_logs(self) -> List[ValidationLog]:
        return self._logs


class ValidationLogs:
    @staticmethod
    def duplicate_feature_names(fqn: str, count: int) -> ValidationLog:
        return ValidationLog(
            header="Duplicate feature names",
            subheader=(
                f'There are {count} features with the same name of "{fqn}". All features require a ' f"distinct name"
            ),
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def no_pkey(namespace: str) -> ValidationLog:
        return ValidationLog(
            header=f"Feature set '{namespace}' is missing a primary feature",
            subheader=f"Please add an 'int' or 'str' feature to '{namespace}', annotated with '= feature(primary=True)'",
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def multiple_pkeys(namespace: str, names: List[str]) -> ValidationLog:
        names_joined = ", ".join(names)
        return ValidationLog(
            header=f"Feature set '{namespace}' has too many primary features",
            subheader=f"Found primary features: {names_joined}. Composite primary keys are not supported. Please mark only a single feature as primary.",
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def max_staleness_expected_with_etl_to_online(fqn: str) -> ValidationLog:
        return ValidationLog(
            header=f'Missing max staleness for "{fqn}"',
            subheader=f'The feature "{fqn}" is set to ETL to online, but doesn\'t specify a max staleness. Any ETL to online would be immediately invalidated.',
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def feature_name_protected(fqn: str) -> ValidationLog:
        return ValidationLog(
            header="Feature uses protected name",
            subheader=f"Feature names cannot begin with '_chalk' or '__'. Found '{fqn}'",
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def namespace_protected(namespace: str, name: str) -> ValidationLog:
        return ValidationLog(
            header="Feature class uses protected namespace",
            subheader=(
                f'The feature "{namespace}.{name}" belongs to the protected namespace "{namespace}". '
                f'Feature namespaces cannot begin with "_chalk" or "__". Please rename this feature set.'
            ),
        )

    @staticmethod
    def duplicate_resolver_name(count: int, name: str) -> ValidationLog:
        return ValidationLog(
            header="Duplicate resolver names",
            subheader=(
                f'There are {count} resolvers with the same name of "{name}". All resolvers require a ' f"distinct name"
            ),
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def multiple_input_namespaces(fqn: str, namespaces: List[str], resolver_definition: str) -> ValidationLog:
        return ValidationLog(
            header=f'Resolver "{fqn}" requires features from multiple namespaces.',
            subheader=f"""All inputs to a resolver should be rooted in the same namespace.
The resolver "{fqn}" takes inputs in the namespaces {oxford_comma_list(add_quotes(namespaces))}.

If you require features from many feature classes, reference them via their relationships, such as:
  User.bank_account.title,
  User.full_name

{resolver_definition}
""",
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def no_outputs(fqn: str) -> ValidationLog:
        return ValidationLog(
            header=f'Resolver "{fqn}" does not define outputs.',
            subheader="See https://docs.chalk.ai/docs/python-resolvers#outputs for information about valid resolver outputs.",
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def multiple_df_outputs(fqn: str) -> ValidationLog:
        return ValidationLog(
            header=f'Resolver "{fqn}" defines multiple DataFrames as output.',
            subheader="See https://docs.chalk.ai/docs/python-resolvers#outputs for information about valid resolver outputs.",
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def mixed_relationship_and_scalar_outputs(fqn: str) -> ValidationLog:
        return ValidationLog(
            header=f'Resolver "{fqn}" returns both relationships and scalar features.',
            subheader="See https://docs.chalk.ai/docs/python-resolvers#outputs for information about valid resolver outputs.",
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def invalid_feature_time_input(resolver_fqn: str, resolver_kind: ResolverKind, feature_fqn: str) -> ValidationLog:
        kind = RESOLVER_ENUM_TO_KIND.get(resolver_kind, "unknown")
        return ValidationLog(
            header=f"Feature times cannot be accepted as input to {kind} resolvers.",
            subheader=(
                f'The resolver "{resolver_fqn}" takes as input the\n'
                f'feature "{feature_fqn}", a feature_time() type.\n'
                f"Feature times can be returned from {kind} resolvers to\n"
                "indicate data from the past, but cannot be accepted\n"
                "as inputs."
            ),
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def unrecognized_input(resolver_fqn: str, feature_fqn: str) -> ValidationLog:
        return ValidationLog(
            header=f"Resolver '{resolver_fqn}' requires an unknown feature '{feature_fqn}'",
            subheader="Please check the inputs to this resolver, and make sure they are all features.",
            severity=export_pb2.VALIDATION_LOG_SEVERITY_ERROR,
        )

    @staticmethod
    def invalid_severity(severity: str, metadata_name: str) -> ValidationLog:
        severities_lower = [e.lower() for e in GraphLogSeverity]
        severity_choices = '" or "'.join(severities_lower)
        return ValidationLog(
            header=f'Found invalid log severity "{severity}" config for missing metadata',
            subheader=(
                f'The required feature metadata "{metadata_name}" is associated with an invalid log severity "{severity}".'
                f' Please use "{severity_choices}" in chalk.yml'
            ),
            severity=export_pb2.VALIDATION_LOG_SEVERITY_WARNING,
        )

    @staticmethod
    def invalid_metadata_name(metadata_name: str, entity: str) -> ValidationLog:
        return ValidationLog(
            header=f'Found invalid feature metadata "{metadata_name}" in config',
            subheader=(
                f'The required metadata "{metadata_name}" is not a valid {entity} metadata.'
                f" Please consider removing it from chalk.yml"
            ),
            severity=export_pb2.VALIDATION_LOG_SEVERITY_WARNING,
        )


def _validate_primary_key(
    feature: graph_pb.FeatureType,
    builder: _LogBuilder,
    lsp_builder: FeatureClassErrorBuilder,
):
    if not feature.HasField("scalar") or not feature.scalar.is_primary:
        return

    fqn = get_feature_type_fqn(feature)
    if not feature.scalar.HasField("arrow_type"):
        raise ValueError(f"The serialized dtype is not set for feature '{fqn}'")
    pa_dtype = FeatureConverter.convert_proto_dtype_to_pa_dtype(feature.scalar.arrow_type)
    if (
        not pa.types.is_integer(pa_dtype)
        and not pa.types.is_large_string(pa_dtype)
        and not pa.types.is_string(pa_dtype)
    ):
        attr_name = get_feature_type_attribute_name(feature)
        lsp_builder.add_diagnostic(
            label="invalid type",
            message=f"Primary keys must be integers or strings. Feature '{attr_name}' has a type of '{pa_dtype}'",
            range=lsp_builder.annotation_range(attr_name),
            code="23",
        )
        builder.add_error(
            header=f'Invalid primary key type "{fqn}"',
            subheader=(f"Primary keys must be integers or strings. " f'Feature "{fqn}" has a type of {pa_dtype}'),
        )


def _validate_etl_offline_to_online(
    feature_set: graph_pb.FeatureSet,
    feature: graph_pb.FeatureType,
    builder: _LogBuilder,
):
    if (feature.HasField("scalar") and feature.scalar.is_primary) or feature.HasField("feature_time"):
        return

    # We want to resolve this before do validate
    # in case `max_staleness` or `etl_offline_to_online`
    # is set on the feature set, but not on individual
    # features.
    eoto = get_feature_type_etl_offline_to_online(feature)
    if eoto is None:
        eoto = feature_set.etl_offline_to_online if feature_set.HasField("etl_offline_to_online") else None

    max_staleness = get_feature_type_max_staleness(feature)
    if max_staleness is None:
        max_staleness = feature_set.max_staleness_duration if feature_set.HasField("max_staleness_duration") else None

    if eoto and max_staleness is None:
        fqn = get_feature_type_fqn(feature)
        builder.add(ValidationLogs.max_staleness_expected_with_etl_to_online(fqn))


def _validate_no_feature_times_as_input(
    fqn_to_feature: Mapping[str, graph_pb.FeatureType],
    resolver: graph_pb.Resolver,
    builder: _LogBuilder,
    lsp_builder: ResolverErrorBuilder,
):
    for i, inp in enumerate(resolver.inputs):
        if not inp.HasField("feature"):
            continue
        if inp.feature.feature.namespace == PSEUDONAMESPACE:
            continue
        fqn = get_feature_reference_fqn(inp.feature.feature)
        if fqn not in fqn_to_feature:
            # We might not ever come in here because `convert_graph` should have validated this.
            lsp_builder.add_diagnostic(
                message=f"Resolver '{resolver.fqn}' requires an unrecognized feature '{fqn}'",
                code="135",
                range=lsp_builder.function_arg_annotation_by_index(i),
                label="invalid resolver input",
            )
            builder.add(ValidationLogs.unrecognized_input(resolver.fqn, fqn))
        elif fqn_to_feature[fqn].HasField("feature_time") and resolver.kind != graph_pb.RESOLVER_KIND_OFFLINE:
            lsp_builder.add_diagnostic(
                message=(
                    f'The resolver "{resolver.fqn}" takes as input the\n'
                    f'feature "{fqn}", a feature_time() type.\n'
                    f"Feature times can be returned from {resolver.kind} resolvers to\n"
                    "indicate data from the past, but cannot be accepted\n"
                    "as inputs."
                ),
                code="136",
                range=lsp_builder.function_arg_annotation_by_index(i),
                label="invalid resolver input",
            )
            builder.add(ValidationLogs.invalid_feature_time_input(resolver.fqn, resolver.kind, fqn))


def _validate_feature_names_unique(features: List[graph_pb.FeatureType], builder: _LogBuilder):
    counter: DefaultDict[Tuple[str, str], int] = defaultdict(lambda: 0)
    for r in features:
        # Need to also index on `attribute_name` because protos store the features generated
        # for versioned features with the same name but different attribute names.
        # CHA-2703 might change the need for this handling.
        counter[get_feature_type_fqn(r), get_feature_type_attribute_name(r)] += 1

    for fqn_and_attr_name, count in counter.items():
        fqn, _ = fqn_and_attr_name
        if count > 1:
            builder.add(ValidationLogs.duplicate_feature_names(fqn, count))


def _validate_resolver_names_unique(resolvers: List[graph_pb.Resolver], builder: _LogBuilder):
    counter: DefaultDict[str, int] = defaultdict(lambda: 0)
    for r in resolvers:
        counter[r.fqn.split(".")[-1]] += 1

    for name, count in counter.items():
        if count > 1:
            builder.add(ValidationLogs.duplicate_resolver_name(count, name))


def _validate_resolver_input(
    singleton_namespaces: set[str],
    resolver: graph_pb.Resolver,
    builder: _LogBuilder,
    lsp_builder: ResolverErrorBuilder,
):
    namespaces = []
    ns_to_arg_num: Dict[str, int] = {}

    for idx, input_ in enumerate(resolver.inputs):
        if input_.HasField("feature"):
            f = input_.feature.feature
            namespace = f.namespace
            if len(f.path) > 0:
                namespace = f.path[0].namespace

            if namespace == PSEUDONAMESPACE:
                continue

            if namespace not in singleton_namespaces:
                ns_to_arg_num[namespace] = idx
                if namespace not in namespaces:
                    namespaces.append(namespace)
        elif input_.HasField("df"):
            # TODO: Validate dataframe
            continue
        elif input_.HasField("state"):
            continue
        else:
            raise ValueError(f"Unknown input type: {input_}")

    if len(namespaces) > 1:
        arg_num = list(ns_to_arg_num.values())[0]
        d = lsp_builder.add_diagnostic(
            message=f"""All inputs to a resolver should be rooted in the same namespace.
The resolver "{resolver.fqn}" takes inputs in the namespaces {oxford_comma_list(add_quotes(namespaces))}.

If you require features from many feature classes, reference them via their relationships, such as:
  User.bank_account.title,
  User.full_name""",
            code="131",
            range=lsp_builder.function_arg_annotation_by_index(arg_num),
            label="secondary namespace",
        )
        for ns, arg in ns_to_arg_num.items():
            if arg != arg_num:
                d.with_range(
                    lsp_builder.function_arg_annotation_by_index(arg),
                    label=f"other namespace {ns}",
                )

        builder.add(
            ValidationLogs.multiple_input_namespaces(
                fqn=resolver.fqn,
                namespaces=namespaces,
                resolver_definition=resolver.function.function_definition,
            )
        )


def _validate_resolver_output(resolver: graph_pb.Resolver, builder: _LogBuilder, lsp_builder: ResolverErrorBuilder):
    for o in resolver.outputs:
        if not o.HasField("df") and not o.HasField("feature"):
            raise ValueError(f"Unexpected output type: {o}")

    features: List[graph_pb.FeatureReference] = [o.feature for o in resolver.outputs if o.HasField("feature")]
    dataframes: List[graph_pb.DataFrameType] = [o.df for o in resolver.outputs if o.HasField("df")]

    if len(features) == 0 and len(dataframes) == 0:
        lsp_builder.add_diagnostic(
            message=f'Resolver "{resolver.fqn}" does not define outputs. All resolvers must have an output',
            code="132",
            range=lsp_builder.function_return_annotation(),
            label="invalid resolver output",
            code_href="https://docs.chalk.ai/docs/python-resolvers#outputs",
        )
        builder.add(ValidationLogs.no_outputs(resolver.fqn))
    if len(dataframes) > 1:
        lsp_builder.add_diagnostic(
            message=f'Resolver "{resolver.fqn}" defines multiple DataFrames as output. At most one is permitted',
            code="133",
            range=lsp_builder.function_return_annotation(),
            label="invalid resolver output",
            code_href="https://docs.chalk.ai/docs/python-resolvers#outputs",
        )
        builder.add(ValidationLogs.multiple_df_outputs(resolver.fqn))
    if len(features) > 0 and len(dataframes) > 0:
        lsp_builder.add_diagnostic(
            message=(
                f'Resolver "{resolver.fqn}" returns both relationships and scalar features. '
                f"Only one or the other is permitted."
            ),
            code="134",
            range=lsp_builder.function_return_annotation(),
            label="invalid resolver output",
            code_href="https://docs.chalk.ai/docs/python-resolvers#outputs",
        )
        builder.add(ValidationLogs.mixed_relationship_and_scalar_outputs(resolver.fqn))


def _validate_feature_names(feature: graph_pb.FeatureType, builder: _LogBuilder, lsp_builder: FeatureClassErrorBuilder):
    namespace = get_feature_type_namespace(feature)
    if namespace == PSEUDONAMESPACE:
        return
    name = get_feature_type_name(feature)
    if name.startswith("__") or name.startswith("_chalk"):
        lsp_builder.add_diagnostic(
            message="Feature names cannot begin with '_chalk' or '__'.",
            range=lsp_builder.property_range(get_feature_type_attribute_name(feature)),
            label="protected name",
            code="24",
        )
        builder.add(ValidationLogs.feature_name_protected(f"{namespace}.{name}"))
    if namespace.startswith("__") or namespace.startswith("_chalk"):
        lsp_builder.add_diagnostic(
            message="Feature classes cannot have names that begin with '_chalk' or '__'.",
            label="protected namespace",
            range=lsp_builder.decorator_kwarg_value_range("name") or lsp_builder.class_definition_range(),
            code="25",
        )
        builder.add(ValidationLogs.namespace_protected(namespace, name))


def _get_field_value(
    entity: Union[graph_pb.FeatureType, graph_pb.Resolver],
    field: ProtoField,
) -> Optional[Any]:
    def get_field_with_presence_considered(
        underlying: Union[
            graph_pb.ScalarFeatureType,
            graph_pb.HasOneFeatureType,
            graph_pb.HasManyFeatureType,
            graph_pb.FeatureTimeFeatureType,
            graph_pb.WindowedFeatureType,
            graph_pb.Resolver,
        ],
    ):
        if not hasattr(underlying, field.name):
            # Important to raise AttributeError first before checking whether `underlying.HasField`
            # so that we raise AttributeError instead of whatever error `HasField` would raise
            raise AttributeError(f"Entity {underlying} does not have attribute {field.name}")
        if field.has_presence and not underlying.HasField(field.name):
            return None
        return getattr(underlying, field.name)

    if isinstance(entity, graph_pb.FeatureType):
        feature = getattr(entity, entity.WhichOneof("type"))
        return get_field_with_presence_considered(feature)
    elif isinstance(entity, graph_pb.Resolver):  # pyright: ignore[reportUnnecessaryIsInstance]
        return get_field_with_presence_considered(entity)
    else:
        raise ValueError(f"Unknown entity type: {entity}")


def _get_missing_metadata(
    entity: Union[graph_pb.FeatureType, graph_pb.Resolver],
    attribute_map: Dict[str, ProtoField],
    entity_type: Literal["feature", "resolver"],
    settings: Collection[MetadataSettings],
    builder: _LogBuilder,
) -> MissingMetadata:
    if isinstance(entity, graph_pb.FeatureType) and entity.HasField("windowed"):
        # TODO: We store windowed features metadata in the pseudofeatures.
        #       instead of the base feature. This might change with CHA-3194.
        return {}

    res: MissingMetadata = collections.defaultdict(list)
    for setting in settings:
        metadata_name = setting.name
        try:
            severity = GraphLogSeverity(setting.missing.upper())
        except:
            """If this fails, _validate_metadata_config should log the error"""
            continue
        if metadata_name in attribute_map:
            attribute_name = attribute_map[metadata_name]
            value = None

            try:
                value = _get_field_value(entity, attribute_name)
            except AttributeError:
                builder.add_warning(
                    header=f'Found invalid {entity_type} metadata "{metadata_name}" in config',
                    subheader=(
                        f'The required metadata "{metadata_name}" is not a valid {entity_type} metadata.'
                        f" Please consider removing it from chalk.yml"
                    ),
                )
            if not value:
                res[severity].append(metadata_name)
    return res


def _get_missing_metadata_features_log(
    missing_metadata_features: List[MissingMetadataFeature], severity: GraphLogSeverity
) -> Union[Tuple[str, str], None]:
    if not missing_metadata_features:
        return None

    missing_messages = []
    feature_str = "feature"
    missing_metadata_header_str = "missing metadata"
    max_name_len = max([len(get_feature_type_name(w.feature)) for w in missing_metadata_features])
    feature_column_width = max(max_name_len, len(feature_str)) + 1
    get_padding = lambda s: feature_column_width - len(s)

    first_feature = missing_metadata_features[0].feature
    header = f'"{to_camel_case(get_feature_type_namespace(first_feature))}" features have missing metadata'
    subheader = f"  Filepath: {missing_metadata_features[0].namespace_path}\n\n"
    subheader = ""

    for wrapper in missing_metadata_features:
        missing_metadata = wrapper.missing_metadata.get(severity)
        if not missing_metadata:
            continue
        padding_1 = get_padding(get_feature_type_name(wrapper.feature))
        missing_metadata_str = ", ".join(missing_metadata)
        missing_messages.append(
            f"      {get_feature_type_name(wrapper.feature)}{' ' * padding_1}: {missing_metadata_str}"
        )

    if not missing_messages:
        return None

    padding_2 = get_padding(feature_str)
    subheader += f"      {'-' * len(feature_str)}{' ' * padding_2}  {'-' * len(missing_metadata_header_str)}\n"
    subheader += f"      {feature_str}{' ' * padding_2}  {missing_metadata_header_str}\n"
    subheader += f"      {'-' * len(feature_str)}{' ' * padding_2}  {'-' * len(missing_metadata_header_str)}\n"
    subheader += "\n".join(missing_messages)

    return header, subheader


def _get_missing_metadata_resolver_log(
    missing_metadata_resolver: MissingMetadataResolver, severity: GraphLogSeverity
) -> Union[Tuple[str, str], None]:
    missing_metadata = missing_metadata_resolver.missing_metadata.get(severity)
    if missing_metadata is None:
        return
    missing_metadata_str = ", ".join(missing_metadata)
    resolver_str = "resolver"
    missing_metadata_header_str = "missing metadata"
    resolver = missing_metadata_resolver.resolver
    max_name_len = len(resolver.fqn)
    feature_column_width = max(max_name_len, len(resolver_str)) + 1
    get_padding = lambda s: feature_column_width - len(s)
    padding_1 = get_padding(resolver.fqn)

    header = f'Resolver "{resolver.fqn}" has missing metadata'
    subheader = f"  Filepath: {missing_metadata_resolver.resolver.function.file_name}\n\n"
    message = f"      {resolver.fqn}{' ' * padding_1}: {missing_metadata_str}"

    padding_2 = get_padding(resolver_str)
    subheader += f"      {'-' * len(resolver_str)}{' ' * padding_2}  {'-' * len(missing_metadata_header_str)}\n"
    subheader += f"      {resolver_str}{' ' * padding_2}  {missing_metadata_header_str}\n"
    subheader += f"      {'-' * len(resolver_str)}{' ' * padding_2}  {'-' * len(missing_metadata_header_str)}\n"
    subheader += message

    return header, subheader


def _build_feature_log_for_each_severity(missing_metadata_features: List[MissingMetadataFeature], builder: _LogBuilder):
    if not missing_metadata_features:
        return

    for severity in GraphLogSeverity:
        log_or_none = _get_missing_metadata_features_log(
            missing_metadata_features=missing_metadata_features, severity=GraphLogSeverity(severity)
        )
        if log_or_none is None:
            continue
        header, subheader = log_or_none
        builder.add_log(header=header, subheader=subheader, severity=GraphLogSeverity(severity))


def _build_resolver_log_for_each_severity(
    missing_metadata_resolvers: List[MissingMetadataResolver], builder: _LogBuilder
):
    if not missing_metadata_resolvers:
        return

    for severity in GraphLogSeverity:
        for resolver in missing_metadata_resolvers:
            log_or_none = _get_missing_metadata_resolver_log(
                missing_metadata_resolver=resolver, severity=GraphLogSeverity(severity)
            )
            if log_or_none is None:
                continue
            header, subheader = log_or_none
            builder.add_log(header=header, subheader=subheader, severity=GraphLogSeverity(severity))


def _validate_feature_metadata(
    namespace_features: Collection[graph_pb.FeatureType],
    namespace_path: str,
    settings: ValidationSettings,
    builder: _LogBuilder,
) -> List[MissingMetadataFeature]:
    wrapped_features = []
    for nf in namespace_features:
        missing_metadata = _get_missing_metadata(
            entity=nf,
            attribute_map=_METADATA_STRING_TO_FEATURE_ATTRIBUTE_MAP,
            entity_type="feature",
            settings=settings.feature.metadata,
            builder=builder,
        )
        if not missing_metadata:
            continue
        wrapped_features.append(
            MissingMetadataFeature(feature=nf, missing_metadata=missing_metadata, namespace_path=namespace_path)
        )

    return wrapped_features


def _validate_metadata(
    feature_sets: List[graph_pb.FeatureSet],
    resolvers: List[graph_pb.Resolver],
    builder: _LogBuilder,
    settings: ValidationSettings,
):
    _validate_metadata_config(builder, settings)

    if settings.HasField("feature") and settings.feature.metadata:
        for fs in feature_sets:
            wf = _validate_feature_metadata(
                namespace_features=fs.features, namespace_path=fs.class_path, settings=settings, builder=builder
            )
            _build_feature_log_for_each_severity(missing_metadata_features=wf, builder=builder)
    if settings.resolver and settings.resolver.metadata:
        missing_resolver_metadatas: List[MissingMetadataResolver] = []
        for r in resolvers:
            metadata = _get_missing_metadata(
                r,
                _METADATA_STRING_TO_RESOLVER_ATTRIBUTE_MAP,
                "resolver",
                settings=settings.resolver.metadata,
                builder=builder,
            )
            missing_resolver_metadatas.append(MissingMetadataResolver(resolver=r, missing_metadata=metadata))
        _build_resolver_log_for_each_severity(missing_resolver_metadatas, builder=builder)


def _validate_metadata_config(builder: _LogBuilder, settings: ValidationSettings):
    if settings.HasField("feature"):
        for missing_metadata in settings.feature.metadata:
            metadata_name = missing_metadata.name
            severity = missing_metadata.missing
            try:
                GraphLogSeverity(severity.upper())
            except ValueError:
                builder.add(ValidationLogs.invalid_severity(severity, metadata_name))
            if metadata_name not in _METADATA_STRING_TO_FEATURE_ATTRIBUTE_MAP:
                builder.add(ValidationLogs.invalid_metadata_name(metadata_name, "feature"))
    if settings.HasField("resolver"):
        for missing_metadata in settings.resolver.metadata:
            metadata_name = missing_metadata.name
            severity = missing_metadata.missing
            try:
                GraphLogSeverity(severity.upper())
            except ValueError:
                builder.add(ValidationLogs.invalid_severity(severity, metadata_name))
            if metadata_name not in _METADATA_STRING_TO_RESOLVER_ATTRIBUTE_MAP:
                builder.add(ValidationLogs.invalid_metadata_name(metadata_name, "resolver"))


def _validate_namespace_primary_key(
    namespace: str,
    features: List[graph_pb.FeatureType],
    builder: _LogBuilder,
    singleton_namespaces: set[str],
):
    if namespace in singleton_namespaces:
        return

    primary_features = list(f for f in features if f.HasField("scalar") and f.scalar.is_primary)

    if len(primary_features) == 0:
        builder.add(ValidationLogs.no_pkey(namespace))
    elif len(primary_features) > 1:
        builder.add(ValidationLogs.multiple_pkeys(namespace, [get_feature_type_name(f) for f in primary_features]))


def validate_artifacts(graph: graph_pb.Graph, config: Optional[ProjectSettings]) -> List[ValidationLog]:
    """Beware: errors here will only be acted upon `chalk apply --branch`.
    Regular deployments will have their graph errors acted upon by the server.
    """
    new_graph = copy.deepcopy(graph)
    for fs in new_graph.feature_sets:
        all_features = list(fs.features)
        fs.ClearField("features")
        fs.features.extend([f for f in all_features if not get_feature_type_is_autogenerated(f)])
    graph = new_graph

    builder = _LogBuilder()
    singleton_namespaces = {c.name for c in graph.feature_sets if c.is_singleton}
    features_flattened = [f for fs in graph.feature_sets for f in fs.features]
    fqn_to_feature = {get_feature_type_fqn(f): f for f in features_flattened}

    # Validate the features
    _validate_feature_names_unique(features_flattened, builder)
    if config is not None and config.HasField("validation"):
        _validate_metadata(list(graph.feature_sets), list(graph.resolvers), builder, config.validation)

    garbage_feature_builder = FeatureClassErrorBuilder(uri="garbage.py", namespace="garbage", node=None)
    for fs in graph.feature_sets:
        _validate_namespace_primary_key(
            namespace=fs.name,
            features=list(fs.features),
            builder=builder,
            singleton_namespaces=singleton_namespaces,
        )
        lsp_builder = (
            FeatureSetBase.registry[fs.name].__chalk_error_builder__
            if fs.name in FeatureSetBase.registry and LSPErrorBuilder.lsp
            else garbage_feature_builder
        )
        for f in fs.features:
            _validate_primary_key(f, builder, lsp_builder)
            _validate_etl_offline_to_online(fs, f, builder)
            _validate_feature_names(f, builder, lsp_builder)

        # Validate joins
        if (fs_class := FeatureSetBase.registry.get(fs.name)) is not None:
            for f_field in fs_class.features:
                try:
                    f_field.join  # (this has the side effect of validating the join)
                except Exception as e:
                    if not LSPErrorBuilder.promote_exception(e):
                        raise

    # Validate the resolvers
    _validate_resolver_names_unique(list(graph.resolvers), builder)
    garbage_resolver_builder = ResolverErrorBuilder(fn=None)
    for resolver in graph.resolvers:
        try:
            lsp_builder = garbage_resolver_builder
            if LSPErrorBuilder.lsp:
                maybe_resolver = RESOLVER_REGISTRY.get_resolver(resolver.fqn)
                if maybe_resolver is not None:
                    if maybe_resolver.is_sql_file_resolver:
                        continue
                    lsp_builder = maybe_resolver.lsp_builder
            _validate_resolver_input(
                singleton_namespaces=singleton_namespaces,
                resolver=resolver,
                builder=builder,
                lsp_builder=lsp_builder,
            )
            _validate_resolver_output(resolver, builder, lsp_builder)
            _validate_no_feature_times_as_input(
                fqn_to_feature=fqn_to_feature, resolver=resolver, builder=builder, lsp_builder=lsp_builder
            )
        except Exception:
            err = traceback.format_exc()
            builder.add_error(
                header=f'Failed to validate resolver "{resolver.fqn}"',
                subheader=f"Please check the resolver for syntax errors.\nRaw error: {err}",
            )

    return builder.get_logs()



from __future__ import annotations

import weakref
from typing import Generic, TypeVar

T = TypeVar("T")


class WeakSetByIdentity(Generic[T]):
    def __init__(self):
        super().__init__()
        self._items: dict[int, weakref.ReferenceType[T]] = {}

    def add(self, item: T):
        """
        Adds the item to this weak set. It will not be kept alive.
        """
        item_id = id(item)

        def remove_item(ref: weakref.ReferenceType[T]):
            if item_id in self._items and self._items[item_id] is ref:
                # We need to confirm that this ref hasn't been replaced by another one,
                # which could happen if the id() was reused by Python right after the `item` was collected.
                del self._items[item_id]

        self._items[item_id] = weakref.ref(item, remove_item)

    def __contains__(self, item: T) -> bool:
        return id(item) in self._items and self._items[id(item)]() is item



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/common/v1/chart.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1b\x63halk/common/v1/chart.proto\x12\x0f\x63halk.common.v1"P\n\x06Series\x12\x16\n\x06points\x18\x01 \x03(\x01R\x06points\x12\x14\n\x05label\x18\x02 \x01(\tR\x05label\x12\x14\n\x05units\x18\x03 \x01(\tR\x05units:\x02\x18\x01"x\n\x05\x43hart\x12\x14\n\x05title\x18\x01 \x01(\tR\x05title\x12/\n\x06series\x18\x02 \x03(\x0b\x32\x17.chalk.common.v1.SeriesR\x06series\x12$\n\x0ex_timestamp_ms\x18\x04 \x03(\x03R\x0cxTimestampMs:\x02\x18\x01\x42\x7f\n\x13\x63om.chalk.common.v1B\nChartProtoP\x01\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43halk.Common.V1\xca\x02\x0f\x43halk\\Common\\V1\xe2\x02\x1b\x43halk\\Common\\V1\\GPBMetadata\xea\x02\x11\x43halk::Common::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.common.v1.chart_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.common.v1B\nChartProtoP\001\242\002\003CCX\252\002\017Chalk.Common.V1\312\002\017Chalk\\Common\\V1\342\002\033Chalk\\Common\\V1\\GPBMetadata\352\002\021Chalk::Common::V1"
    _globals["_SERIES"]._options = None
    _globals["_SERIES"]._serialized_options = b"\030\001"
    _globals["_CHART"]._options = None
    _globals["_CHART"]._serialized_options = b"\030\001"
    _globals["_SERIES"]._serialized_start = 48
    _globals["_SERIES"]._serialized_end = 128
    _globals["_CHART"]._serialized_start = 130
    _globals["_CHART"]._serialized_end = 250
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import dataclasses
import json
import os
import traceback
import uuid
from datetime import datetime, timedelta
from enum import Enum, IntEnum
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
from typing_extensions import TypeAlias

from chalk.byte_transmit.model import ByteBaseModel, ByteDict
from chalk.client._internal_models.models import OfflineQueryGivensVersion
from chalk.features import Feature
from chalk.features._encoding.json import FeatureEncodingOptions
from chalk.features.resolver import Resolver
from chalk.features.tag import EnvironmentId
from chalk.prompts import Prompt
from chalk.queries.query_context import ContextJsonDict
from chalk.utils.df_utils import read_parquet
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl
    from pydantic import BaseModel, Extra, Field, validator

    root_validator = lambda _: (lambda x: x)
else:
    try:
        from pydantic.v1 import BaseModel, Extra, Field, root_validator, validator
    except ImportError:
        from pydantic import BaseModel, Extra, Field, root_validator, validator

MAX_STR_LENGTH = 10_000

CPU_REGEX = r"^(\d+(\.\d+)?m?)$"
MEMORY_REGEX = r"^\d+([EPTGMK]i|[EPTGMk])?$"

TIMEDELTA_PREFIX = "delta:"  # used to disambiguate datetimes and timedeltas in string form

FeatureReference: TypeAlias = Union[str, Any]

_CHALK_DEBUG_FULL_TRACE = os.getenv("CHALK_DEBUG_FULL_TRACE") == "1"


def _category_for_error_code(c: Union[ErrorCode, str]) -> ErrorCodeCategory:
    c = ErrorCode[c]
    return {
        ErrorCode.PARSE_FAILED: ErrorCodeCategory.REQUEST,
        ErrorCode.RESOLVER_NOT_FOUND: ErrorCodeCategory.REQUEST,
        ErrorCode.INVALID_QUERY: ErrorCodeCategory.REQUEST,
        ErrorCode.VALIDATION_FAILED: ErrorCodeCategory.FIELD,
        ErrorCode.RESOLVER_FAILED: ErrorCodeCategory.FIELD,
        ErrorCode.RESOLVER_TIMED_OUT: ErrorCodeCategory.FIELD,
        ErrorCode.UPSTREAM_FAILED: ErrorCodeCategory.FIELD,
        ErrorCode.UNAUTHENTICATED: ErrorCodeCategory.NETWORK,
        ErrorCode.UNAUTHORIZED: ErrorCodeCategory.NETWORK,
        ErrorCode.INTERNAL_SERVER_ERROR: ErrorCodeCategory.NETWORK,
        ErrorCode.CANCELLED: ErrorCodeCategory.NETWORK,
        ErrorCode.DEADLINE_EXCEEDED: ErrorCodeCategory.NETWORK,
    }[c]


class OnlineQueryContext(BaseModel):
    """Context in which to execute a query."""

    environment: Optional[str] = None
    """
    The environment under which to run the resolvers.
    API tokens can be scoped to an # environment.
    If no environment is specified in the query,
    but the token supports only a single environment,
    then that environment will be taken as the scope
    for executing the request.
    """

    tags: Optional[List[str]] = None
    """
    The tags used to scope the resolvers.
    More information at https://docs.chalk.ai/docs/resolver-tags
    """

    required_resolver_tags: Optional[List[str]] = None


class OfflineQueryContext(BaseModel):
    environment: Optional[str] = None
    """
    The environment under which to run the resolvers.
    API tokens can be scoped to an # environment.
    If no environment is specified in the query,
    but the token supports only a single environment,
    then that environment will be taken as the scope
    for executing the request.
    """


class ErrorCode(str, Enum):
    """The detailed error code.

    For a simpler category of error, see `ErrorCodeCategory`.
    """

    PARSE_FAILED = "PARSE_FAILED"
    """The query contained features that do not exist."""

    RESOLVER_NOT_FOUND = "RESOLVER_NOT_FOUND"
    """
    A resolver was required as part of running the dependency
    graph that could not be found.
    """

    INVALID_QUERY = "INVALID_QUERY"
    """
    The query is invalid. All supplied features need to be
    rooted in the same top-level entity.
    """

    VALIDATION_FAILED = "VALIDATION_FAILED"
    """
    A feature value did not match the expected schema
    (e.g. `incompatible type "int"; expected "str"`)
    """

    RESOLVER_FAILED = "RESOLVER_FAILED"
    """The resolver for a feature errored."""

    RESOLVER_TIMED_OUT = "RESOLVER_TIMED_OUT"
    """The resolver for a feature timed out."""

    UPSTREAM_FAILED = "UPSTREAM_FAILED"
    """
    A crash in a resolver that was to produce an input for
    the resolver crashed, and so the resolver could not run
    crashed, and so the resolver could not run.
    """

    UNAUTHENTICATED = "UNAUTHENTICATED"
    """The request was submitted with an invalid authentication header."""

    UNAUTHORIZED = "UNAUTHORIZED"
    """The supplied credentials do not provide the right authorization to execute the request."""

    INTERNAL_SERVER_ERROR = "INTERNAL_SERVER_ERROR"
    """An unspecified error occurred."""

    CANCELLED = "CANCELLED"
    """The operation was cancelled, typically by the caller."""

    DEADLINE_EXCEEDED = "DEADLINE_EXCEEDED"
    """The deadline expired before the operation could complete."""


class ErrorCodeCategory(str, Enum):
    """The category of an error.

    For more detailed error information, see `ErrorCode`
    """

    REQUEST = "REQUEST"
    """
    Request errors are raised before execution of your
    resolver code. They may occur due to invalid feature
    names in the input or a request that cannot be satisfied
    by the resolvers you have defined.
    """

    FIELD = "FIELD"
    """
    Field errors are raised while running a feature resolver
    for a particular field. For this type of error, you'll
    find a feature and resolver attribute in the error type.
    When a feature resolver crashes, you will receive null
    value in the response. To differentiate from a resolver
    returning a null value and a failure in the resolver,
    you need to check the error schema.
    """

    NETWORK = "NETWORK"
    """
    Network errors are thrown outside your resolvers.
    For example, your request was unauthenticated,
    connection failed, or an error occurred within Chalk.
    """


class ChalkException(BaseModel, frozen=True):
    """Information about an exception from a resolver run."""

    kind: str
    """The name of the class of the exception."""

    message: str
    """The message taken from the exception."""

    stacktrace: str
    """The stacktrace produced by the code."""

    internal_stacktrace: Optional[str] = None
    """The stacktrace produced by the code, full detail."""

    @classmethod
    def from_exception(cls, exc: BaseException) -> "ChalkException":
        return ChalkException.create(
            kind=type(exc).__name__,
            message=str(exc),
            stacktrace="".join(traceback.format_exception(exc)),
        )

    @classmethod
    def create(
        cls,
        kind: str,
        message: str,
        stacktrace: str,
        internal_stacktrace: Optional[str] = None,
    ) -> "ChalkException":
        return ChalkException(
            kind=kind,
            message=message[0:MAX_STR_LENGTH],
            stacktrace=stacktrace[-MAX_STR_LENGTH:],
            internal_stacktrace=internal_stacktrace[-MAX_STR_LENGTH:] if internal_stacktrace is not None else None,
        )


class ChalkError(BaseModel, frozen=True):
    """
    The `ChalkError` describes an error from running a resolver
    or from a feature that can't be validated.
    """

    code: ErrorCode
    """The type of the error."""

    category: ErrorCodeCategory = ErrorCodeCategory.NETWORK
    """
    The category of the error, given in the type field for the error codes.
    This will be one of "REQUEST", "NETWORK", and "FIELD".
    """

    message: str
    """A readable description of the error message."""

    display_primary_key: Optional[str] = None
    """
    A human-readable hint that can be used to identify the entity that this error is associated with.
    """

    display_primary_key_fqn: Optional[str] = None
    """
    If provided, can be used to add additional context to 'display_primary_key'.
    """

    exception: Optional[ChalkException] = None
    """The exception that caused the failure, if applicable."""

    feature: Optional[str] = None
    """
    The fully qualified name of the failing feature, e.g. `user.identity.has_voip_phone`.
    """

    resolver: Optional[str] = None
    """
    The fully qualified name of the failing resolver, e.g. `my.project.get_fraud_score`.
    """

    def is_resolver_runtime_error(self) -> bool:
        """
        Returns True if the error indicates an issue with user's resolver, rather than an internal Chalk failure.
        """
        return self.code in (ErrorCode.RESOLVER_FAILED, ErrorCode.RESOLVER_TIMED_OUT, ErrorCode.UPSTREAM_FAILED)

    def copy_for_feature(self, feature: str) -> "ChalkError":
        return self.copy(update={"feature": feature})

    def copy_for_pkey(self, pkey: Union[str, int]) -> "ChalkError":
        return self.copy(update={"display_primary_key": str(pkey)})

    @classmethod
    def create(
        cls,
        code: ErrorCode,
        message: str,
        category: Optional[ErrorCodeCategory] = None,
        display_primary_key: Optional[str] = None,
        display_primary_key_fqn: Optional[str] = None,
        exception: Optional[ChalkException] = None,
        feature: Optional[str] = None,
        resolver: Optional[str] = None,
    ) -> "ChalkError":
        category = category or _category_for_error_code(code)
        if not _CHALK_DEBUG_FULL_TRACE:
            # Truncate the message to a specified maximum length.
            message = message[0:MAX_STR_LENGTH]

        _HAS_CHALK_TRACE = "[has chalk trace]"
        if _CHALK_DEBUG_FULL_TRACE and _HAS_CHALK_TRACE not in message:
            # Include a stack trace if it's not already present and the super-verbose
            # full trace flag is enabled.
            import traceback

            formatted_stack = traceback.format_stack()[:-1]  # Exclude this validation function.
            start_stack_from = 0
            for i in range(len(formatted_stack)):
                if "run_endpoint_function" in formatted_stack[i]:
                    # This function occurs in the stack trace before the actual entry into the engine-
                    # everything before it is boilerplate.
                    start_stack_from = i + 1
            message = (
                f"{message}\n{_HAS_CHALK_TRACE}\n"
                + "[" * 200
                + "\n"
                + "\n".join(formatted_stack[start_stack_from:])
                + "\n"
                + "]" * 200
                + "\n"
            )

        return ChalkError(
            code=code,
            category=category,
            message=message,
            display_primary_key=display_primary_key,
            display_primary_key_fqn=display_primary_key_fqn,
            exception=exception,
            feature=feature,
            resolver=resolver,
        )

    if TYPE_CHECKING:
        # Defining __hash__ only when type checking
        # since pydantic provides a hash for frozen models
        def __hash__(self) -> int:
            ...


class ResolverRunStatus(str, Enum):
    """Status of a scheduled resolver run."""

    RECEIVED = "received"
    """The request to run the resolver has been received, and is running or scheduled."""

    SUCCEEDED = "succeeded"
    """The resolver run failed."""

    FAILED = "failed"
    """The resolver run succeeded."""


class ResolverRunResponse(BaseModel):
    """Status of a scheduled resolver run."""

    id: str
    """The ID of the resolver run."""

    status: ResolverRunStatus
    """The current status of the resolver run."""


class WhoAmIResponse(BaseModel):
    """Response for checking the authenticated user."""

    user: str
    """The ID of the user or service token making the query."""

    environment_id: Optional[str] = None
    """The environment under which the client's queries will be run, unless overridden"""

    team_id: Optional[str] = None
    """The team ID pertaining to the client"""


class FeatureResolutionMeta(BaseModel, frozen=True):
    """Detailed metadata about the execution of an online query."""

    chosen_resolver_fqn: str
    """The name of the resolver that computed the feature value."""

    cache_hit: bool
    """Whether the feature request was satisfied by a cached value."""

    primitive_type: Optional[str] = None
    """
    Primitive type name for the feature, e.g. `str` for `some_feature: str`.
    Returned only if query-level 'include_meta' is True.
    """

    version: int = 1
    """
    The version that was selected for this feature. Defaults to `default_version`, if query
    does not specify a constraint. If no versioning information is provided on the feature definition,
    the default version is `1`.
    """


class FeatureResult(BaseModel):
    field: str
    """
    The name of the feature requested, e.g. 'user.identity.has_voip_phone'.
    """

    value: Any  # Value should be a TJSON type
    """
    The value of the requested feature.
    If an error was encountered in resolving this feature,
    this field will be empty.
    """

    pkey: Any = None
    """The primary key of the resolved feature."""

    error: Optional[ChalkError] = None
    """
    The error code encountered in resolving this feature.
    If no error occurred, this field is empty.
    """

    valid: Optional[bool] = None
    """
    Whether the feature was resolved successfully.
    """

    ts: Optional[datetime] = None
    """
    The time at which this feature was computed.
    This value could be significantly in the past if you're using caching.
    """

    meta: Optional[FeatureResolutionMeta] = None
    """Detailed information about how this feature was computed."""


class ExchangeCredentialsRequest(BaseModel):
    client_id: str
    client_secret: str
    grant_type: str
    scope: Optional[str] = None


class ExchangeCredentialsResponse(BaseModel):
    access_token: str
    token_type: str
    expires_in: int
    # expires_at: datetime
    api_server: str
    primary_environment: Optional[str] = None
    engines: Optional[Mapping[str, str]] = None


class OfflineQueryInput(BaseModel):
    columns: List[str]
    values: List[List[Any]]  # Values should be of type TJSON


class OnlineQueryRequest(BaseModel):
    inputs: Mapping[str, Any]  # Values should be of type TJSON
    outputs: List[str]
    expression_outputs: List[str] = Field(default_factory=list)
    now: Optional[str] = None  # iso format so we can json
    staleness: Optional[Mapping[str, str]] = None
    context: Optional[OnlineQueryContext] = None
    include_meta: bool = True
    explain: Union[bool, Literal["only"]] = False
    skip_online_storage: Optional[bool] = False
    skip_offline_storage: Optional[bool] = False
    skip_metrics_storage: Optional[bool] = False
    skip_cache_lookups: Optional[bool] = False
    correlation_id: Optional[str] = None
    query_name: Optional[str] = None
    query_name_version: Optional[str] = None
    deployment_id: Optional[str] = None
    branch_id: Optional[str] = None
    meta: Optional[Mapping[str, str]] = None
    store_plan_stages: Optional[bool] = False
    encoding_options: FeatureEncodingOptions = FeatureEncodingOptions()
    planner_options: Optional[Mapping[str, Any]] = None
    query_context: Optional[ContextJsonDict] = None
    value_metrics_tag_by_features: Tuple[str, ...] = ()
    overlay_graph: Optional[str] = None  # base64-encoded chalk.graph.v1.OverlayGraph proto


@dataclasses.dataclass
class OnlineQuery:
    input: Union[Mapping[FeatureReference, Sequence[Any]], Any]
    output: Sequence[str]
    staleness: Optional[Mapping[str, str]] = None
    tags: Optional[Sequence[str]] = None
    required_resolver_tags: Optional[Sequence[str]] = None
    value_metrics_tag_by_features: Sequence[str] = ()
    planner_options: Optional[Mapping[str, Any]] = None


class OnlineQueryManyRequest(BaseModel):
    inputs: Mapping[str, List[Any]]
    outputs: List[str]
    expression_outputs: List[str] = Field(default_factory=list)
    now: Optional[List[str]] = None
    staleness: Optional[Mapping[str, str]] = None
    context: Optional[OnlineQueryContext] = None
    include_meta: bool = True
    explain: bool = False
    skip_online_storage: Optional[bool] = False
    skip_offline_storage: Optional[bool] = False
    skip_metrics_storage: Optional[bool] = False
    skip_cache_lookups: Optional[bool] = False
    correlation_id: Optional[str] = None
    query_name: Optional[str] = None
    query_name_version: Optional[str] = None
    deployment_id: Optional[str] = None
    branch_id: Optional[str] = None
    meta: Optional[Mapping[str, str]] = None
    store_plan_stages: Optional[bool] = False
    query_context: Optional[ContextJsonDict] = None
    encoding_options: FeatureEncodingOptions = FeatureEncodingOptions()
    value_metrics_tag_by_features: Tuple[str, ...] = ()
    overlay_graph: Optional[str] = None


class MultiUploadFeaturesRequest(ByteBaseModel):
    features: List[str]
    table_compression: str
    table_bytes: bytes


class MultiUploadFeaturesResponse(BaseModel):
    operation_id: str
    errors: List[ChalkError]


class PersistenceSettings(BaseModel):
    persist_online_storage: Optional[bool] = None
    persist_offline_storage: Optional[bool] = None


class TriggerResolverRunRequest(BaseModel):
    resolver_fqn: str
    upper_bound: Optional[str] = None
    lower_bound: Optional[str] = None
    timestamping_mode: Literal["feature_time", "online_store_write_time"] = "feature_time"
    persistence_settings: Optional[PersistenceSettings] = None
    override_target_image_tag: Optional[str] = None
    idempotency_key: Optional[str] = None


class QueryMeta(BaseModel, extra=Extra.allow):
    execution_duration_s: float
    """
    The time, expressed in seconds, that Chalk spent executing this query.
    """

    deployment_id: Optional[str] = None
    """
    The id of the deployment that served this query.
    """

    environment_id: Optional[str] = None
    """
    The id of the environment that served this query. Not intended to be human readable, but helpful for support.
    """

    environment_name: Optional[str] = None
    """
    The short name of the environment that served this query. For example: "dev" or "prod".
    """

    query_id: Optional[str] = None
    """
    A unique ID generated and persisted by Chalk for this query. All computed features, metrics, and logs are
    associated with this ID. Your system can store this ID for audit and debugging workflows.
    """

    query_timestamp: Optional[datetime] = None
    """
    At the start of query execution, Chalk computes 'datetime.now()'. This value is used to timestamp computed features.
    """

    query_hash: Optional[str] = None
    """
    Deterministic hash of the 'structure' of the query. Queries that have the same input/output features will
    typically have the same hash; changes may be observed over time as we adjust implementation details.
    """

    explain_output: Optional[str] = None
    """
    An unstructured string containing diagnostic information about the query execution. Only included if `explain` is True.
    """


class OnlineQueryResponse(BaseModel):
    data: List[FeatureResult]
    errors: Optional[List[ChalkError]] = None
    meta: Optional[QueryMeta] = None

    def for_fqn(self, fqn: str):
        return next((x for x in self.data if x.field == fqn), None)

    class Config:
        json_encoders = {
            np.integer: int,
            np.floating: float,
        }


@dataclasses.dataclass
class BulkUploadFeaturesResult:
    errors: List[ChalkError]
    trace_id: Optional[str] = None


@dataclasses.dataclass
class BulkOnlineQueryResult:
    """
    Represents the result of a single `OnlineQuery`, returned by
    `query_bulk`.

    The `scalars_df` member holds the primary results of the `OnlineQuery`.
    Access this data using `to_polars()` or `to_pandas()`.
    """

    scalars_df: Optional[pl.DataFrame]
    groups_dfs: Optional[Dict[str, pl.DataFrame]]  # change to chalk df whenever we figure out pydantic
    errors: Optional[List[ChalkError]]
    meta: Optional[QueryMeta]
    trace_id: Optional[str] = None
    """Chalk Support can use this trace ID to investigate a query if an internal error occurs."""

    def to_polars(self) -> pl.DataFrame:
        """
        Allows access to the results of an `OnlineQuery` submitted to `query_bulk` as a Polars dataframe.
        """
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")

        return self.scalars_df if self.scalars_df is not None else pl.DataFrame()

    def to_pandas(self) -> pd.DataFrame:
        """
        Allows access to the results of an `OnlineQuery` submitted to `query_bulk` as a Pandas dataframe.
        """
        if self.scalars_df is not None:
            return self.scalars_df.to_pandas()
        else:
            import pandas as pd

            return pd.DataFrame()

    # def get_feature_value(self, pkey: str | int, f: FeatureReference):
    #     f_casted = ensure_feature(f)
    #     if f_casted.is_has_many:
    #         return self.groups_dfs[f_casted.root_fqn]
    #     else:
    #         return self.scalars_df[0][f_casted.root_fqn].item()


@dataclasses.dataclass
class BulkOnlineQueryResponse:
    results: List[BulkOnlineQueryResult]
    global_errors: List[ChalkError] = dataclasses.field(default_factory=list)
    """Errors that don't correspond to a specific individual query."""

    trace_id: Optional[str] = None
    """Chalk Support can use this trace ID to investigate a query if an internal error occurs."""

    def __getitem__(self, item: int):
        """
        Support `client.query_bulk(...)[0]` syntax.
        """
        return self.results[item]


class SpineSqlRequest(BaseModel):
    sql_query: str


class OfflineQueryInputUri(BaseModel):
    """
    Offline query input that was uploaded to cloud storage by the user.
    We need to figure out how to shard this data when creating jobs.
    The uris aren't in any bucket.
    """

    parquet_uri: str
    """A list of global uris of the sharded parquet files"""

    start_row: Optional[int] = None
    """The start row to read the parquet file from"""

    end_row: Optional[int] = None
    """The end row to read the parquet file from"""

    is_iceberg: bool = False
    """"If this is `True`, this represents an Iceberg source"""

    iceberg_snapshot_id: Optional[int] = None
    """The snapshot_id to use if parquet_uri points to an Iceberg metadata file"""

    iceberg_start_partition: Optional[int] = None
    """The index of the first partition file to load"""

    iceberg_end_partition: Optional[int] = None
    """One-after the index of the last partition file to load"""

    iceberg_filter: Optional[str] = None
    """A SQL filter string applied to the Iceberg source table."""

    aws_role_arn: Optional[str] = None
    """An AWS role arn to assume while performing Iceberg read operations"""

    aws_region: Optional[str] = None
    """The AWS region to query while performing Iceberg read operations"""

    column_name_to_feature_name: Optional[dict[str, Union[str, list[str]]]] = None
    """An optional remapping from column name to (namespace-qualified) Chalk feature names"""


def OfflineQueryIcebergUri(
    *,
    iceberg_uri: str,
    snapshot_id: Optional[int] = None,
    filter: Optional[str] = None,
    aws_role_arn: Optional[str] = None,
    aws_region: Optional[str] = None,
    column_name_to_feature_name: Optional[dict[str, Union[str, list[str]]]] = None,
    start_partition: Optional[int] = None,
    end_partition: Optional[int] = None,
):
    """
    Creates a query input for an Iceberg table.

    - `iceberg_uri`: The URI of a Glue table `glue://<account_id>/<database>/<table_name>` or the S3
      URI of an Iceberg metadata file `s3://example_bucket/example_table/00001-abcdef123.metadata.json`.

    - `snapshot_id`: The snapshot ID of the Iceberg table to read. If `None`, uses the current snapshot.

    - `filter`: A SQL filter expression applied as a filter to the rows of the Iceberg table.
    Iceberg filters can contain logical operations (AND, OR, NOT) and comparison between columns and constants.

    - `aws_role_arn`: A role to assume before accessing the Iceberg table.

    - `aws_region`: The region to load the Iceberg table from.

    - `column_name_to_feature_name`: A mapping from Iceberg column name to Chalk feature name.
    A column can be mapped to a single feature: `user_id: str(User.id)` or to multiple, e.g.
    `created_time: [str(User.created_time), "__ts__"]` to provide an input time column.
    """
    return OfflineQueryInputUri(
        is_iceberg=True,
        parquet_uri=iceberg_uri,
        iceberg_snapshot_id=snapshot_id,
        iceberg_start_partition=start_partition,
        iceberg_end_partition=end_partition,
        iceberg_filter=filter,
        aws_role_arn=aws_role_arn,
        aws_region=aws_region,
        column_name_to_feature_name=column_name_to_feature_name,
    )


class UploadedParquetShardedOfflineQueryInput(BaseModel):
    """
    Offline query input that is sharded parquet files uploaded to cloud storage.
    """

    filenames: Tuple[str, ...] = Field(description="A list of filenames of the sharded parquet files")
    version: OfflineQueryGivensVersion = Field(description="Version of how the inputs is represented in a table")


def _matches_pattern(pattern: str):
    return Field(pattern=pattern, regex=pattern)


class ResourceRequests(BaseModel):
    """
    Override resource requests for processes with isolated resources, e.g., offline queries and cron jobs.
    Note that making these too large could prevent your job from being scheduled, so please test
    before using these in a recurring pipeline.
    """

    cpu: Optional[str] = _matches_pattern(r"^(\d+(\.\d+)?m?)$")
    """
    CPU requests: Increasing this will make some Chalk operations that are parallel and CPU-bound faster.
    Default unit is physical CPU cores, i.e. "8" means 8 CPU cores, "0.5" means half of a CPU core.
    An alternative unit is "millicore", which is one-thousandth of a CPU core, i.e. 500m is half of a CPU core.
    """

    memory: Optional[str] = _matches_pattern(MEMORY_REGEX)
    """
    Memory requests: you can use these to give your pod more memory, i.e. to prevent especially large jobs from OOMing.
    Default unit is bytes, i.e. 1000000000 is 1 gigabyte of memory.
    You can also specify a suffix such as K, M, or G for kilobytes, megabytes, and gigabytes, respectively.
    It's also possible to use the power of two equivalents, such as Ki, Mi, and Gi.
    """

    ephemeral_volume_size: Optional[str] = _matches_pattern(MEMORY_REGEX)
    """Chalk can use this for spilling intermediate state of some large computations, i.e.
    joins, aggregations, and sorting.
    Default unit is bytes, i.e. 1000000000 is 1 gigabyte of memory.
    You can also specify a suffix such as K, M, or G for kilobytes, megabytes, and gigabytes, respectively.
    It's also possible to use the power of two equivalents, such as Ki, Mi, and Gi.
    """

    ephemeral_storage: Optional[str] = _matches_pattern(MEMORY_REGEX)
    """Ephemeral storage for miscellaneous file system access.
    Should probably not be below 1Gi to ensure there's enough space for the Docker image, etc.
    Should also not be too high or else the pod will not be scheduled.
    """

    resource_group: Optional[str] = None
    """Resource group to use for this job. If not specified, the default resource group will be used."""


class CreateOfflineQueryJobRequest(BaseModel):
    output: List[str]
    """A list of output feature root fqns to query"""

    output_expressions: List[str] = Field(default_factory=list)
    """A list of underscore expressions to compute as query outputs, encoded as
    b64-serialized `FeatureExpression` protos"""

    required_output: List[str] = Field(default_factory=list)
    """A list of required output feature root fqns"""

    required_output_expressions: List[str] = Field(default_factory=list)
    """A list of required underscore expressions feature root fqns, , encoded as b64-serialized FeatureExpression protos"""

    destination_format: str
    """The desired output format. Should be 'CSV' or 'PARQUET'"""

    job_id: Optional[uuid.UUID] = None
    """A unique job id. If not specified, one will be auto generated by the server. If specified by the client,
    then jobs with the same ID will be rejected."""

    input: Union[
        OfflineQueryInput,
        Tuple[OfflineQueryInput, ...],
        None,
        UploadedParquetShardedOfflineQueryInput,
        OfflineQueryInputUri,
    ] = None
    """Any givens"""

    max_samples: Optional[int] = None
    """The maximum number of samples. If None, no limit"""

    # Defaults to ``OFFLINE_QUERY_MAX_CACHE_AGE_SECS`` in the chalkengine config
    max_cache_age_secs: Optional[int] = None
    """The maximum staleness, in seconds, for how old the view on the offline store can be.
    That is, data ingested within this interval will not be reflected in this offline query.
    Set to ``0`` to ignore the cache. If not specified, it defaults to 30 minutes.
    """

    observed_at_lower_bound: Optional[str] = None
    """The lower bound for the observed at timestamp (inclusive). If not specified, defaults to the beginning of time"""

    observed_at_upper_bound: Optional[str] = None
    """The upper bound for the observed at timestamp (inclusive). If not specified, defaults to the end of time."""

    dataset_name: Optional[str] = None
    branch: Optional[str] = None
    recompute_features: Union[bool, List[str]] = False
    sample_features: Optional[List[str]] = None
    store_plan_stages: bool = False
    explain: Union[bool, Literal["only"]] = False  # only is deprecated, leaving for backcompat
    tags: Optional[List[str]] = None
    required_resolver_tags: Optional[List[str]] = None
    """
    If specified, all resolvers invoked as part of this query must be tagged with all of these tags.
    Can be used to ensure that expensive resolvers are not executed.
    """

    correlation_id: Optional[str] = None
    query_context: Optional[ContextJsonDict] = None
    planner_options: Optional[Mapping[str, Any]] = None
    use_multiple_computers: bool = False

    spine_sql_query: Optional[Union[str, SpineSqlRequest]] = None

    recompute_request_revision_id: Optional[str] = None
    resources: Optional[ResourceRequests] = None
    env_overrides: Optional[Dict[str, str]] = None

    override_target_image_tag: Optional[str] = None
    enable_profiling: bool = False
    """Enable engine profiling for the job"""

    store_online: bool = False
    store_offline: bool = False

    num_shards: Optional[int] = None
    num_workers: Optional[int] = None
    feature_for_lower_upper_bound: Optional[str] = None

    completion_deadline: Optional[str] = None
    max_retries: Optional[int] = None

    use_job_queue: bool = False

    overlay_graph: Optional[str] = None

    @root_validator
    def _validate_multiple_computers(cls, values: Dict[str, Any]):
        if values["input"] is None or isinstance(
            values["input"], (UploadedParquetShardedOfflineQueryInput, OfflineQueryInputUri)
        ):
            return values
        expected_use_multiple_computers = isinstance(values["input"], tuple)
        if values["use_multiple_computers"] != expected_use_multiple_computers:
            raise ValueError("input should be tuple or uploaded shards exactly when use_multiple_computers is True")
        return values


class OutputExpression(BaseModel):
    base64_proto: str
    python_repr: str


class CreatePromptEvaluationRequest(CreateOfflineQueryJobRequest):
    prompts: List[Union[Prompt, str]]
    dataset_id: Optional[str] = None
    dataset_revision_id: Optional[str] = None
    reference_output: Optional[Union[OutputExpression, str]] = None
    evaluators: Optional[List[str]] = None
    related_named_prompt_ids: Optional[List[str]] = None
    related_evaluation_ids: Optional[List[str]] = None
    meta: Optional[Mapping[str, str]] = None


class ComputeResolverOutputRequest(BaseModel):
    input: OfflineQueryInput
    resolver_fqn: str
    branch: Optional[str] = None
    environment: Optional[str] = None


class DatasetJobStatusRequest(BaseModel):
    job_id: Optional[str]  # same as revision_id
    dataset_id: Optional[str] = None
    dataset_name: Optional[str] = None
    ignore_errors: bool = False
    query_inputs: bool = False


class DatasetRecomputeRequest(BaseModel):
    dataset_name: Optional[str] = None
    branch: str
    dataset_id: Optional[str] = None
    revision_id: Optional[str] = None
    features: List[str]


class RecomputeResolverOutputRequest(BaseModel):
    persistent_id: str
    resolver_fqn: str
    branch: Optional[str] = None
    environment: Optional[str] = None


class ComputeResolverOutputResponse(BaseModel):
    job_id: str
    persistent_id: str
    errors: Optional[List[ChalkError]] = None


class OfflineQueryRequest(BaseModel):
    """V1 OfflineQueryRequest. Not used by the current Chalk Client."""

    output: List[str]  # output features which can be null
    input: Optional[OfflineQueryInput] = None
    dataset: Optional[str] = None
    resources: Optional[ResourceRequests] = None
    max_samples: Optional[int] = None
    max_cache_age_secs: Optional[int] = None
    required_outputs: List[str] = Field(default_factory=list)  # output features which cannot be null


class OfflineQueryResponse(BaseModel):
    """V1 OfflineQueryResponse. Not used by the current Chalk Client."""

    columns: List[str]
    output: List[List[Any]]  # values should be of TJSON types
    errors: Optional[List[ChalkError]] = None


class CreateOfflineQueryJobResponse(BaseModel):
    """
    Attributes:
        job_id: A job ID, which can be used to retrieve the results.
    """

    job_id: uuid.UUID
    version: int = 1  # Field is deprecated
    errors: Optional[List[ChalkError]] = None


class CreateBranchResponse(BaseModel):
    branch_already_exists: bool
    errors: Optional[List[ChalkError]] = None


class ColumnMetadata(BaseModel):
    """This entire model is deprecated."""

    feature_fqn: str
    """The root FQN of the feature for a column"""

    column_name: str
    """The name of the column that corresponds to this feature"""

    dtype: str
    """The data type for this feature"""
    # This field is currently a JSON-stringified version of the SerializeDType property
    # Using a string instead of a pydantic model the SerializedDType encoding does not affect
    # the api layer


class GetOfflineQueryJobResponse(BaseModel):
    is_finished: bool
    """Whether the export job is finished (it runs asynchronously)"""

    version: int = 1  # Backwards compatibility
    """Version number representing the format of the data. The client uses this version number
    to properly decode and load the query results into DataFrames."""

    urls: List[str]
    """A list of short-lived, authenticated URLs that the client can download to retrieve the exported data."""

    errors: Optional[List[ChalkError]] = None

    # deprecated
    columns: Optional[List[ColumnMetadata]] = None
    """Expected columns for the dataframe, including data type information"""


class QueryStatus(IntEnum):
    PENDING_SUBMISSION = 1
    """Pending submission to the database."""

    SUBMITTED = 2
    """Submitted to the database, but not yet running."""

    RUNNING = 3
    """Running in the database."""

    ERROR = 4
    """Error with either submitting or running the job."""

    EXPIRED = 5
    """The job did not complete before an expiration deadline, so there are no results."""

    CANCELLED = 6
    """Manually cancelled before it errored or finished successfully."""

    SUCCESSFUL = 7  #
    """Successfully ran the job."""


class DatasetSampleFilter(BaseModel):
    lower_bound: Union[datetime, timedelta, None] = None
    upper_bound: Union[datetime, timedelta, None] = None
    max_samples: Optional[int] = None


class DatasetFilter(BaseModel):
    sample_filters: DatasetSampleFilter = Field(default_factory=DatasetSampleFilter)
    max_cache_age_secs: Optional[float] = None


class DatasetPartitionResponse(BaseModel):
    performance_summary: Optional[str] = None


class DatasetRevisionResponse(BaseModel):
    dataset_name: Optional[str] = None
    dataset_id: Optional[uuid.UUID] = None
    environment_id: EnvironmentId
    revision_id: Optional[uuid.UUID] = None  # Currently, the revision ID is the job ID that created the revision
    creator_id: str
    outputs: List[str]
    created_at: Optional[datetime] = None
    started_at: Optional[datetime] = None
    terminated_at: Optional[datetime] = None
    givens_uri: Optional[str] = None
    status: QueryStatus
    filters: DatasetFilter
    num_partitions: Optional[int] = None
    partitions: Optional[List[DatasetPartitionResponse]] = None
    num_bytes: Optional[int] = None
    output_uris: str
    output_version: int
    branch: Optional[str] = None
    dashboard_url: Optional[str] = None
    num_computers: int = 1
    errors: Optional[List[ChalkError]] = None
    metadata: Optional[Mapping[str, Any]] = None


class SetDatasetRevisionMetadataRequest(BaseModel):
    metadata: Mapping[str, Any]


class SetDatasetRevisionMetadataResponse(BaseModel):
    revision_id: str
    errors: Optional[List[ChalkError]] = None


class DatasetRecomputeResponse(DatasetRevisionResponse):
    num_computers: Literal[1] = 1  # pyright: ignore[reportIncompatibleVariableOverride]

    @classmethod
    def from_revision_response(
        cls, revision: DatasetRevisionResponse, errors: Optional[List[ChalkError]] = None
    ) -> "DatasetRecomputeResponse":
        return cls(
            revision_id=revision.revision_id,
            environment_id=revision.environment_id,
            creator_id=revision.creator_id,
            outputs=revision.outputs,
            givens_uri=revision.givens_uri,
            status=revision.status,
            filters=revision.filters,
            num_partitions=revision.num_partitions,
            output_uris=revision.output_uris,
            output_version=revision.output_version,
            num_bytes=revision.num_bytes,
            created_at=revision.created_at,
            started_at=revision.started_at,
            terminated_at=revision.terminated_at,
            dataset_name=revision.dataset_name,
            dataset_id=revision.dataset_id,
            branch=revision.branch,
            dashboard_url=revision.dashboard_url,
            errors=errors,
            num_computers=1,
        )


class DatasetResponse(BaseModel):
    is_finished: bool
    """Whether the export job is finished (it runs asynchronously)"""

    version: int = 1  # Backwards compatibility
    """Version number representing the format of the data. The client uses this version number
    to properly decode and load the query results into DataFrames."""

    environment_id: EnvironmentId
    dataset_id: Optional[uuid.UUID] = None
    dataset_name: Optional[str] = None
    revisions: List[DatasetRevisionResponse]
    errors: Optional[List[ChalkError]] = None


class DatasetRevisionResponseType(str, Enum):
    SUMMARY = "summary"
    """
    A summary of the output table
    """

    PREVIEW = "preview"
    """
    A preview (e.g. df.head()) of the output table
    """


class ShardBatchKey(BaseModel):
    shard_id: int
    """
    The identifier for the shard (or 'pod') on which the offline query was run.
    In standard offline queries, this value is always '0'.
    In asynchronous offline queries, this value is 0-indexed and
    incremented for every shard run on a kube pod.
    """

    batch_id: int
    """
    The identifier for the batch run on a single pod/server.
    In most cases, this value is '0', but offline queries may be multi-batch
    in the case of large primary key spines.
    """

    class Config:
        frozen = True

    def __hash__(self):
        return hash((self.shard_id, self.batch_id))

    def __str__(self):
        return f"{self.shard_id}_{self.batch_id}"

    @classmethod
    def from_str(cls, key_str: str) -> "ShardBatchKey":
        shard_id, batch_id = map(int, key_str.split("_"))
        return cls(shard_id=shard_id, batch_id=batch_id)


def read_parquet_with_shard_batch_columns(key: ShardBatchKey, url: str) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    return read_parquet(url).with_columns(
        [pl.lit(key.shard_id).alias("shard_id"), pl.lit(key.batch_id).alias("batch_id")]
    )


class DatasetRevisionInfoResponse(BaseModel):
    revision_id: uuid.UUID
    type: DatasetRevisionResponseType
    urls: Optional[Mapping[ShardBatchKey, str]] = None
    error: Optional[str] = None

    class Config:
        json_encoders = {ShardBatchKey: lambda v: str(v)}

    @validator("urls", pre=True)  # pyright: ignore[reportUntypedFunctionDecorator]
    def parse_shard_batch_key_map(
        cls, value: Optional[Union[Mapping[str, str], Mapping[ShardBatchKey, str]]]
    ) -> Union[Mapping[ShardBatchKey, str], None]:
        if value is None:
            return None
        return {(ShardBatchKey.from_str(k) if isinstance(k, str) else k): v for k, v in value.items()}

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        d = super().dict(**kwargs)
        if d.get("urls") is not None:
            d["urls"] = {str(k): v for k, v in d["urls"].items()}
        return d

    def json(self, *args: Any, **kwargs: Any) -> str:
        data = self.dict()
        if data.get("revision_id"):
            data["revision_id"] = str(data["revision_id"])
        return json.dumps(data, **kwargs)

    def get_shard_batch_keys(self) -> List[ShardBatchKey]:
        if self.urls is None or len(self.urls) == 0:
            default_err_msg = f"dataset revision '{self.revision_id}' {self.type.value} could not be found"
            raise ValueError(self.error or default_err_msg)
        return list(self.urls.keys())

    def concat_urls(self) -> pl.DataFrame:
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")
        assert self.urls is not None, "check self.urls is not None before this call"
        dfs = [read_parquet_with_shard_batch_columns(key, url) for key, url in self.urls.items()]
        return pl.concat(dfs)

    def to_polars(self, shard_batch_key: Optional[ShardBatchKey] = None) -> pl.LazyFrame:
        raise NotImplementedError("not implemented for base class")

    def to_pandas(self, shard_batch_key: Optional[ShardBatchKey] = None) -> pd.DataFrame:
        raise NotImplementedError("not implemented for base class")


DEFAULT_SHARD_BATCH_KEY = ShardBatchKey(shard_id=0, batch_id=0)


class DatasetRevisionSummaryResponse(DatasetRevisionInfoResponse):
    type: DatasetRevisionResponseType = DatasetRevisionResponseType.SUMMARY

    def to_polars(self, shard_batch_key: Optional[ShardBatchKey] = None) -> pl.LazyFrame:
        if shard_batch_key is None:
            shard_batch_key = DEFAULT_SHARD_BATCH_KEY
        if self.urls is None or len(self.urls) == 0:
            default_err_msg = f"dataset revision '{self.revision_id}' {self.type.value} could not be found"
            raise ValueError(self.error or default_err_msg)
        if shard_batch_key in self.urls:
            return read_parquet_with_shard_batch_columns(shard_batch_key, self.urls[shard_batch_key]).lazy()
        raise ValueError(f"ShardBatchKey {shard_batch_key} not found in return")

    def to_pandas(self, shard_batch_key: Optional[ShardBatchKey] = None) -> pd.DataFrame:
        if shard_batch_key is None:
            shard_batch_key = DEFAULT_SHARD_BATCH_KEY
        if self.urls is None or len(self.urls) == 0:
            default_err_msg = f"dataset revision '{self.revision_id}' {self.type.value} could not be found"
            raise ValueError(self.error or default_err_msg)
        if shard_batch_key in self.urls:
            return read_parquet_with_shard_batch_columns(shard_batch_key, self.urls[shard_batch_key]).to_pandas()
        raise ValueError(f"ShardBatchKey {shard_batch_key} not found in return")


class DatasetRevisionPreviewResponse(DatasetRevisionInfoResponse):
    type: DatasetRevisionResponseType = DatasetRevisionResponseType.PREVIEW

    def to_polars(self, shard_batch_key: Optional[ShardBatchKey] = None) -> pl.LazyFrame:
        if self.urls is None or len(self.urls) == 0:
            default_err_msg = f"dataset revision '{self.revision_id}' {self.type.value} could not be found"
            raise ValueError(self.error or default_err_msg)
        if shard_batch_key is None:
            return self.concat_urls().lazy()
        else:
            if shard_batch_key in self.urls:
                return read_parquet_with_shard_batch_columns(shard_batch_key, self.urls[shard_batch_key]).lazy()
            raise ValueError(f"ShardBatchKey {shard_batch_key} not found in return")

    def to_pandas(self, shard_batch_key: Optional[ShardBatchKey] = None) -> pd.DataFrame:
        if self.urls is None or len(self.urls) == 0:
            default_err_msg = f"dataset revision '{self.revision_id}' {self.type.value} could not be found"
            raise ValueError(self.error or default_err_msg)
        if shard_batch_key is None:
            return self.concat_urls().to_pandas()
        else:
            if shard_batch_key in self.urls:
                return read_parquet_with_shard_batch_columns(shard_batch_key, self.urls[shard_batch_key]).to_pandas()
            raise ValueError(f"ShardBatchKey {shard_batch_key} not found in return")


class SingleEntityUpdate(BaseModel):
    entity_type: Literal["feature", "resolver"]
    entity_fqn: str
    entity_shortname: str

    @classmethod
    def for_resolver(cls, resolver: Resolver) -> "SingleEntityUpdate":
        return cls(
            entity_type="resolver",
            entity_fqn=resolver.fqn,
            entity_shortname=resolver.fqn.split(".")[-1],
        )

    @classmethod
    def for_feature(cls, feature: Feature) -> "SingleEntityUpdate":
        return cls(
            entity_type="feature",
            entity_fqn=feature.fqn,
            entity_shortname=feature.name,
        )


class UpdateGraphEntityResponse(BaseModel):
    """
    Represents the result of live updating a graph entity like a resolver or feature class.
    This may result in multiple individual resolvers/features being updated, e.g. if the user
    adds a new feature class w/ multiple new fields.
    """

    added: Optional[List[SingleEntityUpdate]] = None
    modified: Optional[List[SingleEntityUpdate]] = None
    removed: Optional[List[SingleEntityUpdate]] = None

    errors: Optional[List[ChalkError]] = None


class UpdateResolverResponse(BaseModel):
    updated_fqn: Optional[str] = None
    """The resolver fqn that was updated (may not be the same as the one that was requested)"""

    is_new: Optional[bool] = None
    """Whether a new resolver was created, or if an existing one was replaced"""

    errors: Optional[List[ChalkError]] = None


class FeatureObservationDeletionRequest(BaseModel):
    """
    Represents a request to target particular feature observations for deletion. Note that
    the "features" and "tags" fields are mutually exclusive -- either only one of them is
    specified, or neither is specified, in which case deletion will proceed for all
    features of the primary keys specified.
    """

    namespace: str
    """
    The namespace in which the features targeted for deletion reside.
    """

    features: Optional[List[str]]
    """
    An optional list of the feature names of the features that should be deleted
    for the targeted primary keys. Not specifying this and not specifying the "tags" field
    will result in all features being targeted for deletion for the specified primary keys.
    Note that this parameter and the "tags" parameter are mutually exclusive.
    """

    tags: Optional[List[str]]
    """
    An optional list of tags that specify features that should be targeted for deletion.
    If a feature has a tag in this list, its observations for the primary keys you listed
    will be targeted for deletion. Not specifying this and not specifying the "features"
    field will result in all features being targeted for deletion for the specified primary
    keys. Note that this parameter and the "features" parameter are mutually exclusive.
    """

    primary_keys: List[str]
    """
    The primary keys of the observations that should be targeted for deletion.
    """


class FeatureObservationDeletionResponse(BaseModel):
    """
    Contains ChalkErrors for any failures, if any, that might have occurred when trying
    to delete the features that were requested.
    """

    errors: Optional[List[ChalkError]]


class FeatureDropRequest(BaseModel):
    namespace: str
    """Namespace in which the features targeted for drop reside."""

    features: List[str]
    """Names of the features that should be dropped."""


class FeatureDropResponse(BaseModel):
    """
    Contains ChalkErrors for any failures, if any, that might have occurred when trying
    to drop the features that were requested.
    """

    errors: Optional[List[ChalkError]]


class GetIncrementalProgressResponse(BaseModel):
    """
    Returns information about the current state of an incremental resolver.
    Specifically, the recorded timestamps that the resolver uses to process recent data.
    If both timestamp fields are returned as None, this means the current resolver hasn't
    run yet or hasn't stored any progress data. The next time it runs it will ingest all historical data

    More information at https://docs.chalk.ai/docs/sql#incremental-queries
    """

    environment_id: EnvironmentId

    resolver_fqn: str
    """The fully qualified name of the given resolver."""

    query_name: Optional[str] = None

    max_ingested_timestamp: Optional[datetime]
    """The latest timestamp found in ingested data."""

    last_execution_timestamp: Optional[datetime]
    """The latest timestamp at which the resolver was run. If configured to do so, the
    resolver uses this timestamp instead of max_ingested_timestamp to filter input data.
    If None, this means that this value isn't currently used by this resolver.
    """

    errors: Optional[List[ChalkError]] = None


class SetIncrementalProgressRequest(BaseModel):
    """
    Sets the current state of an incremental resolver, specifically the timestamps it uses
    to filter inputs to only recent data, to the given timestamps.

    More information at https://docs.chalk.ai/docs/sql#incremental-queries
    """

    max_ingested_timestamp: Optional[datetime] = None
    """The latest timestamp found in ingested data.
    Timestamp must have a timezone specified.
    """

    last_execution_timestamp: Optional[datetime] = None
    """The latest time the resolver was run. If configured to do so, the
    resolver uses this timestamp instead of max_ingested_timestamp to filter input data.
    Timestamp must have a timezone specified.
    """


class BranchDeployRequest(BaseModel):
    branch_name: str
    """Name of the branch. If branch does not exist, it will be created."""

    create_only: bool = False
    """If true, tries to create a new branch returns an error if the branch already exists."""

    source_deployment_id: Optional[str] = None
    """Use the given deployment's source on the branch. If None, the latest active deployment will be used."""


class BranchDeployResponse(BaseModel):
    branch_name: str
    new_branch_created: bool = False  # deprecated field, API Server takes care of this logic now.

    source_deployment_id: str
    branch_deployment_id: str

    proto_graph_b64: Optional[str] = None
    proto_export_b64: Optional[str] = None
    errors: List[ChalkError] = Field(default_factory=list)


class BranchStartRequest(BaseModel):
    branch_environment_id: str
    """
    The environment id of the branch to start.
    """


class BranchStartResponse(BaseModel):
    status: Union[Literal["ok"], Literal["error"]]
    message: str


BranchIdParam: TypeAlias = Union[None, str, "ellipsis"]
"""
Type used for the 'branch' parameter in calls to the Chalk Client.
The branch can either be:
 1. A string that is used as the branch name for the request
 2. None, in which case the request is _not_ sent to a branch server,
 3. Ellipsis (...), indicating that the branch name (or lack thereof) is
    inferred from the ChalkClient's current branch.
"""


class StreamResolverTestStatus(str, Enum):
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"


class StreamResolverTestMessagePayload(BaseModel):
    key: Optional[str]
    message_str: Optional[str]
    message_bytes: Optional[str]
    timestamp: Optional[datetime]

    class Config:
        # Add custom encoders to make `datetime` JSON serializable
        json_encoders = {datetime: lambda v: v.isoformat()}  # Converts datetime to ISO 8601 string

    def dict(self, *args: Any, **kwargs: Any):
        original_dict = super().dict(*args, **kwargs)
        encoders = self.__config__.json_encoders or {}
        return {
            key: (encoders[type(value)](value) if type(value) in encoders else value)
            for key, value in original_dict.items()
        }


class StreamResolverTestRequest(BaseModel):
    resolver_fqn: str
    num_messages: Optional[int] = None
    test_messages: Optional[List[StreamResolverTestMessagePayload]] = None
    kafka_auto_offset_reset: Optional[Literal["earliest", "latest"]] = "earliest"


class StreamResolverTestResponse(BaseModel):
    status: StreamResolverTestStatus
    data_uri: Optional[str] = None
    errors: Optional[List[ChalkError]] = None
    message: Optional[str] = None

    @property
    def features(self) -> pl.DataFrame:
        if self.data_uri is None:
            raise ValueError(
                (
                    "Features were not saved to storage. "
                    "Please inspect 'ResolverTestResponse.errors' and 'ResolverTestResponse.message'."
                )
            )
        return read_parquet(self.data_uri)


class FeatherBodyType(str, Enum):
    TABLE = "TABLE"
    RECORD_BATCHES = "RECORD_BATCHES"


class OnlineQueryResultFeather(ByteBaseModel):
    has_data: bool
    scalar_data: bytes
    groups_data: ByteDict
    errors: Optional[List[str]]  # inner str is json of ChalkError to mimic Optional[List[ChalkError]]
    meta: Optional[str]  # inner str is json of QueryMeta to mimic Optional[QueryMeta]


class OnlineQueryResponseFeather(ByteBaseModel):
    query_results_bytes: ByteDict


class ResolverReplayResponse(BaseModel):
    urls: Optional[List[str]] = None
    error: Optional[str] = None


class UploadFeaturesRequest(BaseModel):
    input: Mapping[str, List[Any]]  # Values should be of type List[TJSON]
    preview_deployment_id: Optional[str] = None
    correlation_id: Optional[str] = None
    query_name: Optional[str] = None
    meta: Optional[Mapping[str, str]] = None


class UploadFeaturesResponse(BaseModel):
    errors: List[ChalkError]
    """Errors that occurred during feature upload, if any."""

    trace_id: Optional[str] = None
    """If an internal error occurs, this trace ID can be sent to Chalk Support to help debug the root cause."""


class PlanQueryRequest(BaseModel):
    inputs: List[str]
    outputs: List[str]
    expression_outputs: List[str] = Field(default_factory=list)
    staleness: Optional[Mapping[str, str]] = None
    context: Optional[OnlineQueryContext] = None
    query_name: Optional[str] = None
    query_name_version: Optional[str] = None
    deployment_id: Optional[str] = None
    branch_id: Optional[str] = None
    meta: Optional[Mapping[str, str]] = None
    num_input_rows: Optional[int] = None
    explain: bool = False
    store_plan_stages: bool = False
    encoding_options: FeatureEncodingOptions = FeatureEncodingOptions()


class FeatureSchema(BaseModel):
    fqn: str
    primitive_type: Optional[str]
    rich_type: Optional[str]
    nullable: bool
    pyarrow_dtype: str


class PlanQueryResponse(BaseModel):
    rendered_plan: Optional[str]
    output_schema: List[FeatureSchema]
    errors: List[ChalkError]
    structured_plan: Optional[str] = None


class IngestDatasetRequest(BaseModel):
    revision_id: str
    """The ID of the dataset revision to ingest"""

    branch: Optional[str]
    """The branch to ingest the dataset into"""

    outputs: List[str]
    """The output features to return from the dataset"""

    store_online: bool
    """Whether to store the dataset into the online store"""

    store_offline: bool
    """Whether to store the dataset into the offline store"""

    enable_profiling: bool = False
    """Enable engine profiling for the ingestion"""

    planner_options: Optional[Mapping[str, Any]] = None

    online_timestamping_mode: Optional[str] = None
    explain: bool = False


class AnnotatedSignedUploadURL(BaseModel):
    signed_url: str
    """Signed URLs which can be uploaded to using PUT requests"""

    filename: str
    """Filenames which the signed URLs correspond to"""


class OfflineQueryParquetUploadURLResponse(BaseModel):
    # there is one pair of url for each partition
    urls: Tuple[AnnotatedSignedUploadURL, ...]
    """Signed URLs which can be uploaded to using PUT requests"""


class FeatureStatistics(BaseModel):
    feature_fqn: str
    count: int
    null_count: int
    zero_count: Optional[int]
    mean: Optional[float]
    std: Optional[float]
    max: Optional[float]
    min: Optional[float]
    # each tuple is of the form (percentile, value) where 0 (minimum) <= percentile <= 1 (maximum)
    approx_percentiles: Optional[List[Tuple[float, float]]]
    logical_type: Optional[str]


class FeatureStatisticsResponse(BaseModel):
    data: List[FeatureStatistics]


class PingRequest(BaseModel):
    num: Optional[int] = None


class PingResponse(BaseModel):
    num: int



from __future__ import annotations

import collections.abc
from typing import Any, Collection, Dict, Generic, Iterable, List, Sequence, Set, Tuple, TypeVar, Union, cast, overload

from typing_extensions import Annotated, get_args, get_origin

try:
    from types import UnionType
except ImportError:
    UnionType = Union

_NoneType = type(None)
_AnnotatedType = type(Annotated[int, ""])

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
B_co = TypeVar("B_co", covariant=True)
U = TypeVar("U")


def unwrap_annotated_if_needed(typ: Any):
    args = get_args(typ)
    if type(typ) is _AnnotatedType and hasattr(typ, "__metadata__"):  # pyright: ignore[reportUnnecessaryComparison]
        return args[0]
    return typ


def is_optional(typ: Any) -> bool:
    args = get_args(typ)
    return get_origin(typ) in (Union, UnionType) and len(args) == 2 and any(d is _NoneType for d in args)


def unwrap_optional_and_annotated_if_needed(typ: Any):
    typ = unwrap_annotated_if_needed(typ)
    if is_optional(typ):
        return next(m for m in get_args(typ) if m is not _NoneType)
    return typ


def is_namedtuple(value: Any) -> bool:
    """Infer whether value is a NamedTuple."""
    # From https://github.com/pola-rs/polars/blob/5f3e332fb2a653064f083b02949c527e0ec0afda/py-polars/polars/internals/construction.py#L78
    return all(hasattr(value, attr) for attr in ("_fields", "_field_defaults", "_replace"))


def flatten(v: Sequence[T | Sequence[T | Sequence[T]]]) -> List[T]:
    ret = []
    for x in v:
        if isinstance(x, collections.abc.Sequence) and not isinstance(x, (str, bytes, bytearray)):
            ret.extend(flatten(x))
        else:
            ret.append(x)
    return ret


@overload
def chunks(lst: List[T], n: int) -> Iterable[List[T]]:
    ...


@overload
def chunks(lst: Set[T], n: int) -> Iterable[Set[T]]:
    ...


@overload
def chunks(lst: Tuple[T, ...], n: int) -> Iterable[Tuple[T, ...]]:
    ...


@overload
def chunks(lst: Iterable[T], n: int) -> Iterable[Iterable[T]]:
    ...


def chunks(lst: Iterable[T], n: int) -> Iterable[Iterable[T]]:
    """Yield successive n-sized chunks from ``lst``, potentially lazily.

    Chunks are generated up-front if ``lst`` is a list, tuple, or set. In this case,
    the result will be an iterator of n-sized chunks of the underlying collection type.

    Otherwise, the chunks and the contents of each chunk are generated lazily,
    where at most one sample from ``lst`` is read in advance.

    Parameters
    ----------
    lst
        The collection
    n
        The chunk size. If <=0, then everything will be yielded back as one chunk.
    """
    if n <= 0:
        yield lst

    iterator = iter(lst)

    # Greedily loading the next sample to detect if we exhausted the iterable, so we don't yield an empty chunk at the end
    have_sample_to_yield = True
    try:
        next_sample = next(iterator)
    except StopIteration:
        have_sample_to_yield = False
        return

    def _get_chunk() -> Iterable[T]:
        """Get a chunk of size n from the generator, or raises StopIteration of the lst is empty"""
        nonlocal next_sample, have_sample_to_yield
        for _ in range(n):
            yield next_sample

            try:
                next_sample = next(iterator)
            except StopIteration:
                have_sample_to_yield = False
                break

    while have_sample_to_yield:
        chunk = _get_chunk()
        if isinstance(lst, (tuple, set, list)):
            # If the underlying collection is already allocated, then allocate the entire chunk at once
            # For backwards compatibility
            yield type(lst)(chunk)
        else:
            yield chunk


def ensure_tuple(x: T | Sequence[T] | Dict[Any, T] | None) -> Tuple[T, ...]:
    """Converts ``x`` into a tuple.
    * If ``x`` is `None`, then ``tuple()`` is returned.
    * If ``x`` is a tuple, then ``x`` is returned as-is.
    * If ``x`` is a list, then ``tuple(x)`` is returned.
    * If ``x`` is a dict, then ``tuple(v for v in x.values())`` is returned.
    Otherwise, a single element tuple of ``(x,)`` is returned.

    Parameters
    ----------
    x
        The input to convert into a tuple.

    Returns
    -------
    tuple
        A tuple of ``x``.
    """
    # From https://github.com/mosaicml/composer/blob/020ca02e3848ee8fb6b7fff0c8123f597b05be8a/composer/utils/iter_helpers.py#L40
    if x is None:
        return ()
    if isinstance(x, (str, bytes, bytearray)):
        return (cast(T, x),)
    if isinstance(x, collections.abc.Sequence):
        return tuple(x)
    if isinstance(x, dict):
        return tuple(x.values())
    return (x,)


def get_unique_item(collection: Iterable[T], name: str | None = None) -> T:
    """Get the unique item from the collection, ignoring equal duplicates.

    Raises a ValueError if there is not a single, unique element."""
    unique_item_or_none = get_unique_item_if_exists(collection, name)
    if unique_item_or_none is None:
        in_name = f" in `{name}`" if name is not None else ""
        raise ValueError(f"There should be at least one item{in_name}")
    return unique_item_or_none


def get_unique_item_if_exists(iterable: Iterable[T], name: str | None = None) -> T | None:
    """Get the unique item from the collection, if any element exists inside the collection.

    Raises a ValueError if there are multiple unique elements. Returns None if the collection is empty.
    """
    item = ...
    in_name = f" in `{name}`" if name is not None else ""
    for x in iterable:
        if item is ...:
            item = x
        if x != item:
            raise ValueError(f"Multiple values{in_name} are not permitted. Found {x}, {item}")
    if item is ...:
        return None
    return item


class BidirectionalMap(Generic[T, U]):
    """
    Represents a bijection between two sets of elements, with fast lookup for each direction.
    Under the hood keeps two dictionaries, one from T-->U and another from U-->T
    Constructor enforces that the mapping is a bijection.
    """

    def __init__(self, value_pairs: Sequence[Tuple[T, U]]):
        """
        Takes a sequence of pairs (t,u) to build the mapping t_i <--> u_i. Must be a bijection.
        :param value_pairs:
        """
        super().__init__()
        self._t_to_u: Dict[T, U] = dict(value_pairs)
        self._u_to_t: Dict[U, T] = dict([(v, k) for (k, v) in value_pairs])
        if (len(value_pairs) != len(self._t_to_u)) or (len(value_pairs) != len(self._u_to_t)):
            raise ValueError("Mapping is not a bijection.")

    def get(self, t: T, default: U | None = None) -> U | None:
        return self._t_to_u.get(t, default)

    def contains(self, t: T) -> bool:
        return t in self._t_to_u

    def get_reverse(self, u: U, default: T | None = None) -> T | None:
        return self._u_to_t.get(u, default)

    def contains_reverse(self, u: U) -> bool:
        return u in self._u_to_t

    def insert(self, t: T, u: U):
        """
        Raises if either element already exists.
        """
        if self.contains(t):
            raise ValueError(f"Element {t} already exists in forward mapping.")
        if self.contains_reverse(u):
            raise ValueError(f"Element {t} already exists in reverse mapping.")

    def __len__(self):
        return len(self._t_to_u)

    @classmethod
    def empty(cls):
        return cls(())


class OrderedSet(Generic[T]):
    """
    A_co set of items, with O(1) contains checks, which always iterates in insertion order.
    """

    __slots__ = ("_items",)

    def __init__(self, items: Iterable[T] = ()) -> None:
        """
        Create a new ordered set containing `items`.
        """
        super().__init__()
        self._items: dict[T, None] = dict.fromkeys(items)

    def add(self, item: T) -> None:
        """Adds `item` to the set, if it is not already present."""
        self._items[item] = None

    def update(self, items: Iterable[T]) -> None:
        """Adds all items from `items` to the set, skipping those that are already present."""
        for item in items:
            self._items[item] = None

    def remove(self, item: T) -> None:
        """Removes `item` from the set, if it is present. Otherwise, raises a KeyError."""
        del self._items[item]

    def discard(self, item: T) -> None:
        """Removes `item` from the set, if it is present. Otherwise, does nothing."""
        self._items.pop(item, None)

    def __len__(self) -> int:
        """Returns the number of items in the set."""
        return len(self._items)

    def __iter__(self):
        """Iterates over the items in insertion order."""
        return iter(self._items)

    def __contains__(self, item: object) -> bool:
        """Returns whether the given item is in the set."""
        return item in self._items

    def __or__(self, __value: Iterable[B_co]) -> "OrderedSet[T | B_co]":
        """Creates a new set containing values from both the left and the right."""
        return OrderedSet(tuple(self._items) + tuple(__value))

    def __and__(self, __value: Iterable[T]) -> "OrderedSet[T]":
        """Creates a new set containing only values found in both the left and the right."""
        value_set = set(__value)
        return OrderedSet(item for item in self._items if item in value_set)

    def __sub__(self, __value: Iterable[Any]) -> OrderedSet[T]:
        right = FrozenOrderedSet(__value)
        return OrderedSet(item for item in self if item not in right)

    def __rsub__(self, __value: Iterable[T]) -> Collection[T]:
        if isinstance(__value, (set, frozenset)):
            return __value - set(self)
        raise NotImplementedError

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, (set, frozenset)):
            return frozenset(self._items) == __value
        if not isinstance(__value, (OrderedSet, FrozenOrderedSet)):
            return NotImplemented
        return self._items == __value._items  # pyright: ignore[reportPrivateUsage]

    def issubset(self, other: set[T] | frozenset[T] | OrderedSet[T] | FrozenOrderedSet[T]) -> bool:
        return all(item in other for item in self)

    def issuperset(self, other: Iterable[Any]) -> bool:
        return all(item in self for item in other)

    def isdisjoint(self, other: Iterable[Any]) -> bool:
        return all(item not in self for item in other)

    def freeze(self) -> FrozenOrderedSet[T]:
        return FrozenOrderedSet(self)

    def __repr__(self) -> str:
        return f"OrderedSet({repr(tuple(self._items))})"

    def pop(self) -> T:
        """Removes the item that was last inserted into the ordered set"""
        item, _ = self._items.popitem()
        return item

    def pop_left(self) -> T:
        """Removes the item that was first inserted into the ordered set"""
        item = next(iter(self._items))
        del self._items[item]
        return item

    def clear(self) -> None:
        self._items.clear()


class FrozenOrderedSet(Generic[T_co]):
    """
    Frozen version of `OrderedSet[T_co]`, which can be used as a key in maps and sets.
    """

    __slots__ = ("_items", "_items_frozenset")

    def __init__(self, items: Iterable[T_co] = ()) -> None:
        super().__init__()
        if isinstance(items, FrozenOrderedSet):
            self._items: dict[T_co, None] = items._items
            self._items_frozenset = items._items_frozenset
        else:
            self._items = dict.fromkeys(items)
            self._items_frozenset = frozenset(self._items)

    def __len__(self) -> int:
        """Returns the number of items in the set."""
        return len(self._items)

    def __iter__(self):
        """Iterates over the items in insertion order."""
        return iter(self._items)

    def __contains__(self, item: object) -> bool:
        """Returns whether the given item is in the set."""
        return item in self._items

    def __or__(self, __value: Iterable[B_co]) -> "FrozenOrderedSet[T_co | B_co]":
        """Creates a new set containing values from both the left and the right."""
        return FrozenOrderedSet((*self._items, *__value))

    def __and__(self, __value: Iterable[T_co]) -> "FrozenOrderedSet[T_co]":
        """Creates a new set containing only values found in both the left and the right."""
        other_frozenset = frozenset(__value)
        return FrozenOrderedSet(item for item in self._items if item in other_frozenset)

    def __sub__(self, __value: Iterable[Any]) -> FrozenOrderedSet[T_co]:
        right = frozenset(__value)
        return FrozenOrderedSet(item for item in self if item not in right)

    def __rsub__(self, __value: Iterable[B_co]) -> FrozenOrderedSet[T_co | B_co]:
        return FrozenOrderedSet(x for x in __value if x not in self._items_frozenset)

    def __ror__(self, __value: Iterable[B_co]) -> FrozenOrderedSet[T_co | B_co]:
        return FrozenOrderedSet((*__value, *self._items))

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, FrozenOrderedSet):
            return self._items_frozenset == __value._items_frozenset
        if isinstance(__value, (set, frozenset)):
            return self._items_frozenset == __value
        if isinstance(__value, OrderedSet):
            return self._items_frozenset == frozenset(__value)
        return NotImplemented

    def issubset(
        self,
        other: set[T_co | B_co] | frozenset[T_co | B_co] | OrderedSet[T_co | B_co] | FrozenOrderedSet[T_co | B_co],
    ) -> bool:
        return all(item in other for item in self)

    def issuperset(self, other: Iterable[Any]) -> bool:
        return all(item in self._items_frozenset for item in other)

    def isdisjoint(self, other: Iterable[Any]) -> bool:
        return all(item not in self._items_frozenset for item in other)

    def __hash__(self) -> int:
        return hash(self._items_frozenset)

    def __repr__(self) -> str:
        return f"FrozenOrderedSet({repr(tuple(self._items))})"


def unwrap_optional(x: T | None) -> T:
    if x is None:
        raise ValueError("Value is None")
    return x



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/status.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1c\x63halk/server/v1/status.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1cgoogle/protobuf/struct.proto"\x9c\x03\n\x0bHealthCheck\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12:\n\x06status\x18\x02 \x01(\x0e\x32".chalk.server.v1.HealthCheckStatusR\x06status\x12\x1d\n\x07message\x18\x03 \x01(\tH\x00R\x07message\x88\x01\x01\x12\x38\n\x07latency\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationH\x01R\x07latency\x88\x01\x01\x12\x39\n\tkube_data\x18\x05 \x01(\x0b\x32\x17.google.protobuf.StructH\x02R\x08kubeData\x88\x01\x01\x12\x46\n\x08metadata\x18\x06 \x03(\x0b\x32*.chalk.server.v1.HealthCheck.MetadataEntryR\x08metadata\x1a;\n\rMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\n\n\x08_messageB\n\n\x08_latencyB\x0c\n\n_kube_data"d\n\x12HealthCheckFilters\x12\x12\n\x04name\x18\x01 \x03(\tR\x04name\x12:\n\x06status\x18\x02 \x03(\x0e\x32".chalk.server.v1.HealthCheckStatusR\x06status"d\n\x12\x43heckHealthRequest\x12\x42\n\x07\x66ilters\x18\x01 \x01(\x0b\x32#.chalk.server.v1.HealthCheckFiltersH\x00R\x07\x66ilters\x88\x01\x01\x42\n\n\x08_filters"K\n\x13\x43heckHealthResponse\x12\x34\n\x06\x63hecks\x18\x01 \x03(\x0b\x32\x1c.chalk.server.v1.HealthCheckR\x06\x63hecks"b\n\x10GetHealthRequest\x12\x42\n\x07\x66ilters\x18\x01 \x01(\x0b\x32#.chalk.server.v1.HealthCheckFiltersH\x00R\x07\x66ilters\x88\x01\x01\x42\n\n\x08_filters"I\n\x11GetHealthResponse\x12\x34\n\x06\x63hecks\x18\x01 \x03(\x0b\x32\x1c.chalk.server.v1.HealthCheckR\x06\x63hecks"\x1a\n\x18GetClusterMetricsRequest"5\n\x19GetClusterMetricsResponse\x12\x18\n\x07metrics\x18\x01 \x01(\tR\x07metrics*u\n\x11HealthCheckStatus\x12#\n\x1fHEALTH_CHECK_STATUS_UNSPECIFIED\x10\x00\x12\x1a\n\x16HEALTH_CHECK_STATUS_OK\x10\x01\x12\x1f\n\x1bHEALTH_CHECK_STATUS_FAILING\x10\x02\x32\xc1\x02\n\rHealthService\x12`\n\x0b\x43heckHealth\x12#.chalk.server.v1.CheckHealthRequest\x1a$.chalk.server.v1.CheckHealthResponse"\x06\x90\x02\x01\x80}\x01\x12Z\n\tGetHealth\x12!.chalk.server.v1.GetHealthRequest\x1a".chalk.server.v1.GetHealthResponse"\x06\x90\x02\x01\x80}\x02\x12r\n\x11GetClusterMetrics\x12).chalk.server.v1.GetClusterMetricsRequest\x1a*.chalk.server.v1.GetClusterMetricsResponse"\x06\x90\x02\x01\x80}\x02\x42\x94\x01\n\x13\x63om.chalk.server.v1B\x0bStatusProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.status_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\013StatusProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_HEALTHCHECK_METADATAENTRY"]._options = None
    _globals["_HEALTHCHECK_METADATAENTRY"]._serialized_options = b"8\001"
    _globals["_HEALTHSERVICE"].methods_by_name["CheckHealth"]._options = None
    _globals["_HEALTHSERVICE"].methods_by_name["CheckHealth"]._serialized_options = b"\220\002\001\200}\001"
    _globals["_HEALTHSERVICE"].methods_by_name["GetHealth"]._options = None
    _globals["_HEALTHSERVICE"].methods_by_name["GetHealth"]._serialized_options = b"\220\002\001\200}\002"
    _globals["_HEALTHSERVICE"].methods_by_name["GetClusterMetrics"]._options = None
    _globals["_HEALTHSERVICE"].methods_by_name["GetClusterMetrics"]._serialized_options = b"\220\002\001\200}\002"
    _globals["_HEALTHCHECKSTATUS"]._serialized_start = 1098
    _globals["_HEALTHCHECKSTATUS"]._serialized_end = 1215
    _globals["_HEALTHCHECK"]._serialized_start = 145
    _globals["_HEALTHCHECK"]._serialized_end = 557
    _globals["_HEALTHCHECK_METADATAENTRY"]._serialized_start = 460
    _globals["_HEALTHCHECK_METADATAENTRY"]._serialized_end = 519
    _globals["_HEALTHCHECKFILTERS"]._serialized_start = 559
    _globals["_HEALTHCHECKFILTERS"]._serialized_end = 659
    _globals["_CHECKHEALTHREQUEST"]._serialized_start = 661
    _globals["_CHECKHEALTHREQUEST"]._serialized_end = 761
    _globals["_CHECKHEALTHRESPONSE"]._serialized_start = 763
    _globals["_CHECKHEALTHRESPONSE"]._serialized_end = 838
    _globals["_GETHEALTHREQUEST"]._serialized_start = 840
    _globals["_GETHEALTHREQUEST"]._serialized_end = 938
    _globals["_GETHEALTHRESPONSE"]._serialized_start = 940
    _globals["_GETHEALTHRESPONSE"]._serialized_end = 1013
    _globals["_GETCLUSTERMETRICSREQUEST"]._serialized_start = 1015
    _globals["_GETCLUSTERMETRICSREQUEST"]._serialized_end = 1041
    _globals["_GETCLUSTERMETRICSRESPONSE"]._serialized_start = 1043
    _globals["_GETCLUSTERMETRICSRESPONSE"]._serialized_end = 1096
    _globals["_HEALTHSERVICE"]._serialized_start = 1218
    _globals["_HEALTHSERVICE"]._serialized_end = 1539
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.webhook_pb2 import (
    CreateWebhookRequest,
    CreateWebhookResponse,
    DeleteWebhookRequest,
    DeleteWebhookResponse,
    GetWebhookRequest,
    GetWebhookResponse,
    ListWebhooksRequest,
    ListWebhooksResponse,
    UpdateWebhookRequest,
    UpdateWebhookResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class WebhookServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    CreateWebhook: UnaryUnaryMultiCallable[
        CreateWebhookRequest,
        CreateWebhookResponse,
    ]
    UpdateWebhook: UnaryUnaryMultiCallable[
        UpdateWebhookRequest,
        UpdateWebhookResponse,
    ]
    DeleteWebhook: UnaryUnaryMultiCallable[
        DeleteWebhookRequest,
        DeleteWebhookResponse,
    ]
    GetWebhook: UnaryUnaryMultiCallable[
        GetWebhookRequest,
        GetWebhookResponse,
    ]
    ListWebhooks: UnaryUnaryMultiCallable[
        ListWebhooksRequest,
        ListWebhooksResponse,
    ]

class WebhookServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def CreateWebhook(
        self,
        request: CreateWebhookRequest,
        context: ServicerContext,
    ) -> CreateWebhookResponse: ...
    @abstractmethod
    def UpdateWebhook(
        self,
        request: UpdateWebhookRequest,
        context: ServicerContext,
    ) -> UpdateWebhookResponse: ...
    @abstractmethod
    def DeleteWebhook(
        self,
        request: DeleteWebhookRequest,
        context: ServicerContext,
    ) -> DeleteWebhookResponse: ...
    @abstractmethod
    def GetWebhook(
        self,
        request: GetWebhookRequest,
        context: ServicerContext,
    ) -> GetWebhookResponse: ...
    @abstractmethod
    def ListWebhooks(
        self,
        request: ListWebhooksRequest,
        context: ServicerContext,
    ) -> ListWebhooksResponse: ...

def add_WebhookServiceServicer_to_server(servicer: WebhookServiceServicer, server: Server) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.authtesting_pb2 import (
    GetAdminTestEndpointRequest,
    GetAdminTestEndpointResponse,
    GetAuthedTestEndpointRequest,
    GetAuthedTestEndpointResponse,
    GetDataScientistTestEndpointRequest,
    GetDataScientistTestEndpointResponse,
    GetDeveloperTestEndpointRequest,
    GetDeveloperTestEndpointResponse,
    GetOwnerTestEndpointRequest,
    GetOwnerTestEndpointResponse,
    GetUnauthedTestEndpointRequest,
    GetUnauthedTestEndpointResponse,
    GetViewerTestEndpointRequest,
    GetViewerTestEndpointResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class AuthTestingServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetUnauthedTestEndpoint: UnaryUnaryMultiCallable[
        GetUnauthedTestEndpointRequest,
        GetUnauthedTestEndpointResponse,
    ]
    GetAuthedTestEndpoint: UnaryUnaryMultiCallable[
        GetAuthedTestEndpointRequest,
        GetAuthedTestEndpointResponse,
    ]
    GetViewerTestEndpoint: UnaryUnaryMultiCallable[
        GetViewerTestEndpointRequest,
        GetViewerTestEndpointResponse,
    ]
    GetDataScientistTestEndpoint: UnaryUnaryMultiCallable[
        GetDataScientistTestEndpointRequest,
        GetDataScientistTestEndpointResponse,
    ]
    GetDeveloperTestEndpoint: UnaryUnaryMultiCallable[
        GetDeveloperTestEndpointRequest,
        GetDeveloperTestEndpointResponse,
    ]
    GetAdminTestEndpoint: UnaryUnaryMultiCallable[
        GetAdminTestEndpointRequest,
        GetAdminTestEndpointResponse,
    ]
    GetOwnerTestEndpoint: UnaryUnaryMultiCallable[
        GetOwnerTestEndpointRequest,
        GetOwnerTestEndpointResponse,
    ]

class AuthTestingServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetUnauthedTestEndpoint(
        self,
        request: GetUnauthedTestEndpointRequest,
        context: ServicerContext,
    ) -> GetUnauthedTestEndpointResponse: ...
    @abstractmethod
    def GetAuthedTestEndpoint(
        self,
        request: GetAuthedTestEndpointRequest,
        context: ServicerContext,
    ) -> GetAuthedTestEndpointResponse: ...
    @abstractmethod
    def GetViewerTestEndpoint(
        self,
        request: GetViewerTestEndpointRequest,
        context: ServicerContext,
    ) -> GetViewerTestEndpointResponse: ...
    @abstractmethod
    def GetDataScientistTestEndpoint(
        self,
        request: GetDataScientistTestEndpointRequest,
        context: ServicerContext,
    ) -> GetDataScientistTestEndpointResponse: ...
    @abstractmethod
    def GetDeveloperTestEndpoint(
        self,
        request: GetDeveloperTestEndpointRequest,
        context: ServicerContext,
    ) -> GetDeveloperTestEndpointResponse: ...
    @abstractmethod
    def GetAdminTestEndpoint(
        self,
        request: GetAdminTestEndpointRequest,
        context: ServicerContext,
    ) -> GetAdminTestEndpointResponse: ...
    @abstractmethod
    def GetOwnerTestEndpoint(
        self,
        request: GetOwnerTestEndpointRequest,
        context: ServicerContext,
    ) -> GetOwnerTestEndpointResponse: ...

def add_AuthTestingServiceServicer_to_server(servicer: AuthTestingServiceServicer, server: Server) -> None: ...



from chalk._gen.chalk.kubernetes.v1 import nodes_pb2 as _nodes_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class KarpenterNodepoolDisruptionBudget(_message.Message):
    __slots__ = ("nodes", "schedule", "duration")
    NODES_FIELD_NUMBER: _ClassVar[int]
    SCHEDULE_FIELD_NUMBER: _ClassVar[int]
    DURATION_FIELD_NUMBER: _ClassVar[int]
    nodes: str
    schedule: str
    duration: _duration_pb2.Duration
    def __init__(
        self,
        nodes: _Optional[str] = ...,
        schedule: _Optional[str] = ...,
        duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
    ) -> None: ...

class KarpenterNodepoolDisruption(_message.Message):
    __slots__ = ("consolidate_after", "consolidation_policy", "budgets")
    CONSOLIDATE_AFTER_FIELD_NUMBER: _ClassVar[int]
    CONSOLIDATION_POLICY_FIELD_NUMBER: _ClassVar[int]
    BUDGETS_FIELD_NUMBER: _ClassVar[int]
    consolidate_after: _duration_pb2.Duration
    consolidation_policy: str
    budgets: _containers.RepeatedCompositeFieldContainer[KarpenterNodepoolDisruptionBudget]
    def __init__(
        self,
        consolidate_after: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        consolidation_policy: _Optional[str] = ...,
        budgets: _Optional[_Iterable[_Union[KarpenterNodepoolDisruptionBudget, _Mapping]]] = ...,
    ) -> None: ...

class KarpenterNodeClassRef(_message.Message):
    __slots__ = ("name", "kind", "group")
    NAME_FIELD_NUMBER: _ClassVar[int]
    KIND_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    name: str
    kind: str
    group: str
    def __init__(self, name: _Optional[str] = ..., kind: _Optional[str] = ..., group: _Optional[str] = ...) -> None: ...

class KarpenterNodeSelectorRequirement(_message.Message):
    __slots__ = ("key", "operator", "values")
    KEY_FIELD_NUMBER: _ClassVar[int]
    OPERATOR_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    key: str
    operator: str
    values: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self, key: _Optional[str] = ..., operator: _Optional[str] = ..., values: _Optional[_Iterable[str]] = ...
    ) -> None: ...

class KarpenterNodepoolTemplateSpec(_message.Message):
    __slots__ = (
        "taints",
        "startup_taints",
        "requirements",
        "node_class_ref",
        "expire_after",
        "termination_grace_period",
    )
    TAINTS_FIELD_NUMBER: _ClassVar[int]
    STARTUP_TAINTS_FIELD_NUMBER: _ClassVar[int]
    REQUIREMENTS_FIELD_NUMBER: _ClassVar[int]
    NODE_CLASS_REF_FIELD_NUMBER: _ClassVar[int]
    EXPIRE_AFTER_FIELD_NUMBER: _ClassVar[int]
    TERMINATION_GRACE_PERIOD_FIELD_NUMBER: _ClassVar[int]
    taints: _containers.RepeatedCompositeFieldContainer[_nodes_pb2.KubernetesNodeTaint]
    startup_taints: _containers.RepeatedCompositeFieldContainer[_nodes_pb2.KubernetesNodeTaint]
    requirements: _containers.RepeatedCompositeFieldContainer[KarpenterNodeSelectorRequirement]
    node_class_ref: KarpenterNodeClassRef
    expire_after: _duration_pb2.Duration
    termination_grace_period: _duration_pb2.Duration
    def __init__(
        self,
        taints: _Optional[_Iterable[_Union[_nodes_pb2.KubernetesNodeTaint, _Mapping]]] = ...,
        startup_taints: _Optional[_Iterable[_Union[_nodes_pb2.KubernetesNodeTaint, _Mapping]]] = ...,
        requirements: _Optional[_Iterable[_Union[KarpenterNodeSelectorRequirement, _Mapping]]] = ...,
        node_class_ref: _Optional[_Union[KarpenterNodeClassRef, _Mapping]] = ...,
        expire_after: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        termination_grace_period: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
    ) -> None: ...

class KarpenterNodepoolTemplateMetadata(_message.Message):
    __slots__ = ("labels", "annotations")
    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    class AnnotationsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    LABELS_FIELD_NUMBER: _ClassVar[int]
    ANNOTATIONS_FIELD_NUMBER: _ClassVar[int]
    labels: _containers.ScalarMap[str, str]
    annotations: _containers.ScalarMap[str, str]
    def __init__(
        self, labels: _Optional[_Mapping[str, str]] = ..., annotations: _Optional[_Mapping[str, str]] = ...
    ) -> None: ...

class KarpenterNodepoolTemplate(_message.Message):
    __slots__ = ("metadata", "spec")
    METADATA_FIELD_NUMBER: _ClassVar[int]
    SPEC_FIELD_NUMBER: _ClassVar[int]
    metadata: KarpenterNodepoolTemplateMetadata
    spec: KarpenterNodepoolTemplateSpec
    def __init__(
        self,
        metadata: _Optional[_Union[KarpenterNodepoolTemplateMetadata, _Mapping]] = ...,
        spec: _Optional[_Union[KarpenterNodepoolTemplateSpec, _Mapping]] = ...,
    ) -> None: ...

class KarpenterNodepoolSpec(_message.Message):
    __slots__ = ("disruption", "template", "limits", "weight")
    class LimitsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    DISRUPTION_FIELD_NUMBER: _ClassVar[int]
    TEMPLATE_FIELD_NUMBER: _ClassVar[int]
    LIMITS_FIELD_NUMBER: _ClassVar[int]
    WEIGHT_FIELD_NUMBER: _ClassVar[int]
    disruption: KarpenterNodepoolDisruption
    template: KarpenterNodepoolTemplate
    limits: _containers.ScalarMap[str, str]
    weight: int
    def __init__(
        self,
        disruption: _Optional[_Union[KarpenterNodepoolDisruption, _Mapping]] = ...,
        template: _Optional[_Union[KarpenterNodepoolTemplate, _Mapping]] = ...,
        limits: _Optional[_Mapping[str, str]] = ...,
        weight: _Optional[int] = ...,
    ) -> None: ...

class KarpenterNodepoolMetadata(_message.Message):
    __slots__ = ("annotations", "creation_timestamp", "generation", "name", "resource_version", "uid", "labels")
    class AnnotationsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    ANNOTATIONS_FIELD_NUMBER: _ClassVar[int]
    CREATION_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    GENERATION_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_VERSION_FIELD_NUMBER: _ClassVar[int]
    UID_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    annotations: _containers.ScalarMap[str, str]
    creation_timestamp: _timestamp_pb2.Timestamp
    generation: int
    name: str
    resource_version: str
    uid: str
    labels: _containers.ScalarMap[str, str]
    def __init__(
        self,
        annotations: _Optional[_Mapping[str, str]] = ...,
        creation_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        generation: _Optional[int] = ...,
        name: _Optional[str] = ...,
        resource_version: _Optional[str] = ...,
        uid: _Optional[str] = ...,
        labels: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class KarpenterNodepoolCondition(_message.Message):
    __slots__ = ("type", "status", "last_transition_time", "reason", "message")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    LAST_TRANSITION_TIME_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    type: str
    status: str
    last_transition_time: int
    reason: str
    message: str
    def __init__(
        self,
        type: _Optional[str] = ...,
        status: _Optional[str] = ...,
        last_transition_time: _Optional[int] = ...,
        reason: _Optional[str] = ...,
        message: _Optional[str] = ...,
    ) -> None: ...

class KarpenterNodepoolStatus(_message.Message):
    __slots__ = ("resources", "conditions")
    class ResourcesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    RESOURCES_FIELD_NUMBER: _ClassVar[int]
    CONDITIONS_FIELD_NUMBER: _ClassVar[int]
    resources: _containers.ScalarMap[str, str]
    conditions: _containers.RepeatedCompositeFieldContainer[KarpenterNodepoolCondition]
    def __init__(
        self,
        resources: _Optional[_Mapping[str, str]] = ...,
        conditions: _Optional[_Iterable[_Union[KarpenterNodepoolCondition, _Mapping]]] = ...,
    ) -> None: ...

class KarpenterNodepool(_message.Message):
    __slots__ = ("api_version", "kind", "metadata", "spec", "status")
    API_VERSION_FIELD_NUMBER: _ClassVar[int]
    KIND_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    SPEC_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    api_version: str
    kind: str
    metadata: KarpenterNodepoolMetadata
    spec: KarpenterNodepoolSpec
    status: KarpenterNodepoolStatus
    def __init__(
        self,
        api_version: _Optional[str] = ...,
        kind: _Optional[str] = ...,
        metadata: _Optional[_Union[KarpenterNodepoolMetadata, _Mapping]] = ...,
        spec: _Optional[_Union[KarpenterNodepoolSpec, _Mapping]] = ...,
        status: _Optional[_Union[KarpenterNodepoolStatus, _Mapping]] = ...,
    ) -> None: ...



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.common.v1 import chalk_error_pb2 as _chalk_error_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class ExecuteSqlQueryRequest(_message.Message):
    __slots__ = ("query",)
    QUERY_FIELD_NUMBER: _ClassVar[int]
    query: str
    def __init__(self, query: _Optional[str] = ...) -> None: ...

class ExecuteSqlQueryResponse(_message.Message):
    __slots__ = ("query_id", "parquet", "errors")
    QUERY_ID_FIELD_NUMBER: _ClassVar[int]
    PARQUET_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    query_id: str
    parquet: bytes
    errors: _containers.RepeatedCompositeFieldContainer[_chalk_error_pb2.ChalkError]
    def __init__(
        self,
        query_id: _Optional[str] = ...,
        parquet: _Optional[bytes] = ...,
        errors: _Optional[_Iterable[_Union[_chalk_error_pb2.ChalkError, _Mapping]]] = ...,
    ) -> None: ...

class PlanSqlQueryRequest(_message.Message):
    __slots__ = ("query",)
    QUERY_FIELD_NUMBER: _ClassVar[int]
    query: str
    def __init__(self, query: _Optional[str] = ...) -> None: ...

class PlanSqlQueryResponse(_message.Message):
    __slots__ = ("logical_plan",)
    LOGICAL_PLAN_FIELD_NUMBER: _ClassVar[int]
    logical_plan: str
    def __init__(self, logical_plan: _Optional[str] = ...) -> None: ...



from chalk._gen.chalk.common.v1 import operation_kind_pb2 as _operation_kind_pb2
from chalk._gen.chalk.common.v1 import query_status_pb2 as _query_status_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class VersionedQueryName(_message.Message):
    __slots__ = ("query_name", "query_name_version")
    QUERY_NAME_FIELD_NUMBER: _ClassVar[int]
    QUERY_NAME_VERSION_FIELD_NUMBER: _ClassVar[int]
    query_name: str
    query_name_version: str
    def __init__(self, query_name: _Optional[str] = ..., query_name_version: _Optional[str] = ...) -> None: ...

class QueryLogFilters(_message.Message):
    __slots__ = (
        "operation_id",
        "operation_kind",
        "query_name",
        "agent_id",
        "branch_name",
        "correlation_id",
        "trace_id",
        "query_plan_id",
        "deployment_id",
        "query_status",
        "meta_query_hash",
    )
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    OPERATION_KIND_FIELD_NUMBER: _ClassVar[int]
    QUERY_NAME_FIELD_NUMBER: _ClassVar[int]
    AGENT_ID_FIELD_NUMBER: _ClassVar[int]
    BRANCH_NAME_FIELD_NUMBER: _ClassVar[int]
    CORRELATION_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    QUERY_PLAN_ID_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    QUERY_STATUS_FIELD_NUMBER: _ClassVar[int]
    META_QUERY_HASH_FIELD_NUMBER: _ClassVar[int]
    operation_id: _containers.RepeatedScalarFieldContainer[str]
    operation_kind: _containers.RepeatedScalarFieldContainer[_operation_kind_pb2.OperationKind]
    query_name: _containers.RepeatedCompositeFieldContainer[VersionedQueryName]
    agent_id: _containers.RepeatedScalarFieldContainer[str]
    branch_name: _containers.RepeatedScalarFieldContainer[str]
    correlation_id: _containers.RepeatedScalarFieldContainer[str]
    trace_id: _containers.RepeatedScalarFieldContainer[str]
    query_plan_id: _containers.RepeatedScalarFieldContainer[str]
    deployment_id: _containers.RepeatedScalarFieldContainer[str]
    query_status: _containers.RepeatedScalarFieldContainer[_query_status_pb2.QueryStatus]
    meta_query_hash: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        operation_id: _Optional[_Iterable[str]] = ...,
        operation_kind: _Optional[_Iterable[_Union[_operation_kind_pb2.OperationKind, str]]] = ...,
        query_name: _Optional[_Iterable[_Union[VersionedQueryName, _Mapping]]] = ...,
        agent_id: _Optional[_Iterable[str]] = ...,
        branch_name: _Optional[_Iterable[str]] = ...,
        correlation_id: _Optional[_Iterable[str]] = ...,
        trace_id: _Optional[_Iterable[str]] = ...,
        query_plan_id: _Optional[_Iterable[str]] = ...,
        deployment_id: _Optional[_Iterable[str]] = ...,
        query_status: _Optional[_Iterable[_Union[_query_status_pb2.QueryStatus, str]]] = ...,
        meta_query_hash: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class GetQueryLogEntriesPageToken(_message.Message):
    __slots__ = ("operation_id_hwm", "query_timestamp_hwm")
    OPERATION_ID_HWM_FIELD_NUMBER: _ClassVar[int]
    QUERY_TIMESTAMP_HWM_FIELD_NUMBER: _ClassVar[int]
    operation_id_hwm: str
    query_timestamp_hwm: _timestamp_pb2.Timestamp
    def __init__(
        self,
        operation_id_hwm: _Optional[str] = ...,
        query_timestamp_hwm: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...

class GetQueryLogEntriesRequest(_message.Message):
    __slots__ = (
        "query_timestamp_lower_bound_inclusive",
        "query_timestamp_upper_bound_exclusive",
        "filters",
        "page_size",
        "page_token",
    )
    QUERY_TIMESTAMP_LOWER_BOUND_INCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    QUERY_TIMESTAMP_UPPER_BOUND_EXCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    query_timestamp_lower_bound_inclusive: _timestamp_pb2.Timestamp
    query_timestamp_upper_bound_exclusive: _timestamp_pb2.Timestamp
    filters: QueryLogFilters
    page_size: int
    page_token: str
    def __init__(
        self,
        query_timestamp_lower_bound_inclusive: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        query_timestamp_upper_bound_exclusive: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        filters: _Optional[_Union[QueryLogFilters, _Mapping]] = ...,
        page_size: _Optional[int] = ...,
        page_token: _Optional[str] = ...,
    ) -> None: ...

class QueryLogEntry(_message.Message):
    __slots__ = (
        "operation_id",
        "environment_id",
        "deployment_id",
        "operation_kind",
        "query_timestamp",
        "execution_started_at",
        "execution_finished_at",
        "query_status",
        "query_name",
        "query_name_version",
        "agent_id",
        "branch_name",
        "correlation_id",
        "trace_id",
        "query_plan_id",
        "value_tables",
        "meta_query_hash",
    )
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    OPERATION_KIND_FIELD_NUMBER: _ClassVar[int]
    QUERY_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_STARTED_AT_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_FINISHED_AT_FIELD_NUMBER: _ClassVar[int]
    QUERY_STATUS_FIELD_NUMBER: _ClassVar[int]
    QUERY_NAME_FIELD_NUMBER: _ClassVar[int]
    QUERY_NAME_VERSION_FIELD_NUMBER: _ClassVar[int]
    AGENT_ID_FIELD_NUMBER: _ClassVar[int]
    BRANCH_NAME_FIELD_NUMBER: _ClassVar[int]
    CORRELATION_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    QUERY_PLAN_ID_FIELD_NUMBER: _ClassVar[int]
    VALUE_TABLES_FIELD_NUMBER: _ClassVar[int]
    META_QUERY_HASH_FIELD_NUMBER: _ClassVar[int]
    operation_id: str
    environment_id: str
    deployment_id: str
    operation_kind: _operation_kind_pb2.OperationKind
    query_timestamp: _timestamp_pb2.Timestamp
    execution_started_at: _timestamp_pb2.Timestamp
    execution_finished_at: _timestamp_pb2.Timestamp
    query_status: _query_status_pb2.QueryStatus
    query_name: str
    query_name_version: str
    agent_id: str
    branch_name: str
    correlation_id: str
    trace_id: str
    query_plan_id: str
    value_tables: _containers.RepeatedScalarFieldContainer[str]
    meta_query_hash: str
    def __init__(
        self,
        operation_id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        deployment_id: _Optional[str] = ...,
        operation_kind: _Optional[_Union[_operation_kind_pb2.OperationKind, str]] = ...,
        query_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        execution_started_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        execution_finished_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        query_status: _Optional[_Union[_query_status_pb2.QueryStatus, str]] = ...,
        query_name: _Optional[str] = ...,
        query_name_version: _Optional[str] = ...,
        agent_id: _Optional[str] = ...,
        branch_name: _Optional[str] = ...,
        correlation_id: _Optional[str] = ...,
        trace_id: _Optional[str] = ...,
        query_plan_id: _Optional[str] = ...,
        value_tables: _Optional[_Iterable[str]] = ...,
        meta_query_hash: _Optional[str] = ...,
    ) -> None: ...

class GetQueryLogEntriesResponse(_message.Message):
    __slots__ = ("entries", "next_page_token")
    ENTRIES_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    entries: _containers.RepeatedCompositeFieldContainer[QueryLogEntry]
    next_page_token: str
    def __init__(
        self,
        entries: _Optional[_Iterable[_Union[QueryLogEntry, _Mapping]]] = ...,
        next_page_token: _Optional[str] = ...,
    ) -> None: ...
